import datetime
from unittest import mock

import pytest

import paasta_tools.check_spark_jobs as check_spark_jobs
from paasta_tools.mesos.framework import Framework


@pytest.fixture
def mock_smtp():
    with mock.patch(
        "paasta_tools.check_spark_jobs.smtplib", autospec=True
    ) as mock_smtp:
        yield mock_smtp


@pytest.fixture
def mock_current_time():
    return datetime.datetime(2019, 4, 3, 0, 0, 0)


@pytest.fixture(autouse=True)
def mock_datetime(mock_current_time):
    with mock.patch(
        "paasta_tools.check_spark_jobs.datetime", autospec=True
    ) as mock_time:
        mock_time.datetime.now.return_value = mock_current_time
        mock_time.datetime.fromtimestamp = datetime.datetime.fromtimestamp
        mock_time.timedelta = datetime.timedelta
        yield mock_time


@pytest.fixture
def mock_get_frameworks(mock_current_time):
    mock_current_timestamp = mock_current_time.timestamp()
    mock_frameworks = [
        {
            "id": "uuid1",
            "name": "not_spark_25_hours",
            "user": "test_user",
            "active": True,
            "registered_time": mock_current_timestamp - 60 * 60 * 25,
        },
        {
            "id": "uuid2",
            "name": "spark_2_hours",
            "principal": "spark",
            "user": "test_user",
            "active": True,
            "webui_url": "url2",
            "registered_time": mock_current_timestamp - 60 * 60 * 2,
        },
        {
            "id": "uuid3",
            "name": "spark_25_hours",
            "principal": "spark",
            "user": "test_user",
            "active": True,
            "webui_url": "url3",
            "registered_time": mock_current_timestamp - 60 * 60 * 25,
        },
        {
            "id": "uuid4",
            "name": "spark_25_hours_inactive",
            "principal": "spark",
            "user": "test_user",
            "active": False,
            "webui_url": "url4",
            "registered_time": mock_current_timestamp - 60 * 60 * 25,
        },
    ]
    with mock.patch(
        "paasta_tools.check_spark_jobs.mesos_tools.get_all_frameworks", autospec=True
    ) as mock_get_frameworks:
        mock_get_frameworks.return_value = [
            Framework(config) for config in mock_frameworks
        ]
        yield mock_get_frameworks


@pytest.mark.parametrize(
    "properties,expected",
    [
        (None, None),
        ([["something_else", "value"]], None),
        (
            [["something_else", "value"], ["spark.executorEnv.PAASTA_SERVICE", "foo"]],
            "foo",
        ),
        ([["spark.executorEnv.PAASTA_SERVICE", "jupyterhub_foo"]], "foo"),
    ],
)
def test_guess_service(properties, expected):
    assert check_spark_jobs.guess_service(properties) == expected


@mock.patch("paasta_tools.check_spark_jobs.get_spark_properties", autospec=True)
@mock.patch("paasta_tools.check_spark_jobs.guess_service", autospec=True)
def test_get_matching_framework_info(
    mock_guess_service, mock_get_spark_properties, mock_get_frameworks
):
    mock_get_spark_properties.return_value = None

    more_than_20_hours = check_spark_jobs.get_matching_framework_info(min_hours=20)
    assert len(more_than_20_hours) == 1
    assert more_than_20_hours == [
        {
            "id": "uuid3",
            "name": "spark_25_hours",
            "webui_url": "url3",
            "service": mock_guess_service.return_value,
            "user": "test_user",
            "time_running": "1 day, 1:00:00",
        }
    ]

    more_than_one_hour = check_spark_jobs.get_matching_framework_info(min_hours=1)
    result_names = [result["name"] for result in more_than_one_hour]
    assert result_names == ["spark_2_hours", "spark_25_hours"]


@pytest.mark.parametrize(
    "framework_user,framework_name,expected",
    [
        ("me", "paasta_spark_run_something", "me"),
        ("root", "Custom Spark App", None),
        ("root", "jupyterhub_bill-search-learning_bill_39904_1234", "bill"),
    ],
)
def test_email_user(mock_smtp, framework_user, framework_name, expected):
    info = {
        "id": "id1",
        "name": framework_name,
        "webui_url": "url1",
        "service": "something",
        "user": framework_user,
        "time_running": "1 day, 03:32:00",
    }
    check_spark_jobs.email_user(info, "test.com")

    mock_send_message = mock_smtp.SMTP.return_value.__enter__.return_value.send_message
    if expected:
        assert mock_send_message.call_count == 1
        msg = mock_send_message.call_args[0][0]
        assert msg["To"] == f"{expected}@test.com"
    else:
        assert mock_send_message.call_count == 0


@mock.patch("paasta_tools.check_spark_jobs.email_user", autospec=True)
@mock.patch("paasta_tools.check_spark_jobs.list_services", autospec=True)
@mock.patch("paasta_tools.check_spark_jobs.get_matching_framework_info", autospec=True)
@mock.patch("paasta_tools.check_spark_jobs.update_check_status", autospec=True)
@pytest.mark.parametrize("no_notify", [True, False])
def test_report_spark_jobs(
    mock_check, mock_get_info, mock_list_services, mock_email_user, no_notify
):
    mock_list_services.return_value = ["service1", "service2", "other_service"]
    mock_get_info.return_value = [
        {
            "id": "uuid1",
            "name": "spark1",
            "webui_url": "url1",
            "service": "service1",
            "user": "test_user",
            "time_running": "7:00:00",
        },
        {
            "id": "uuid2",
            "name": "spark2",
            "webui_url": "url2",
            "service": "service2",
            "user": "test_user",
            "time_running": "7:00:00",
        },
        {
            "id": "uuid3",
            "name": "spark3",
            "webui_url": "url3",
            "service": "service2",
            "user": "test_user",
            "time_running": "7:00:00",
        },
        {
            "id": "uuid3",
            "name": "spark3",
            "webui_url": "url3",
            "service": "service_dne",
            "user": "test_user",
            "time_running": "7:00:00",
        },
    ]
    assert check_spark_jobs.report_spark_jobs(1, no_notify, "test.com") == 1
    assert mock_get_info.call_args_list == [mock.call(min_hours=1)]
    if no_notify:
        assert mock_check.call_count == 0
    else:
        assert sorted(mock_check.call_args_list) == sorted(
            [
                mock.call("service1", mock.ANY, 1),
                mock.call("service2", mock.ANY, 1),
                mock.call("other_service", mock.ANY, 0),
            ]
        )
        assert mock_email_user.call_args_list == [
            mock.call(info, "test.com") for info in mock_get_info.return_value
        ]



#!/usr/bin/env python
# Copyright 2015-2017 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import json
import os
import shutil
import subprocess
import sys
import time
from typing import List

from paasta_tools.cli.utils import get_paasta_oapi_api_clustername
from paasta_tools.cli.utils import get_paasta_oapi_client_with_auth
from paasta_tools.cli.utils import lazy_choices_completer
from paasta_tools.cli.utils import run_interactive_cli
from paasta_tools.kubernetes.remote_run import format_remote_run_job_name
from paasta_tools.kubernetes.remote_run import load_eks_or_adhoc_deployment_config
from paasta_tools.kubernetes.remote_run import TOOLBOX_MOCK_SERVICE
from paasta_tools.paastaapi.exceptions import ApiException
from paasta_tools.paastaapi.model.remote_run_start import RemoteRunStart
from paasta_tools.paastaapi.model.remote_run_stop import RemoteRunStop
from paasta_tools.utils import get_username
from paasta_tools.utils import list_all_instances_for_service
from paasta_tools.utils import list_clusters
from paasta_tools.utils import list_services
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import SystemPaastaConfig


KUBECTL_EXEC_CMD_TEMPLATE = (
    "{kubectl_wrapper} --token {token} exec -it -n {namespace} {pod} -- /bin/bash"
)
KUBECTL_CP_TO_CMD_TEMPLATE = (
    "{kubectl_wrapper} --token {token} -n {namespace} cp {source} {pod}:{dest}"
)
KUBECTL_CP_FROM_CMD_TEMPLATE = (
    "{kubectl_wrapper} --token {token} -n {namespace} cp {pod}:{source} {dest}"
)
KUBECTL_LOGS_CMD_TEMPLATE = (
    "{kubectl_wrapper} --token {token} logs -n {namespace} -f {pod}"
)


def _list_services_and_toolboxes() -> List[str]:
    try:
        toolbox_instances = list_all_instances_for_service(
            TOOLBOX_MOCK_SERVICE, instance_type="adhoc"
        )
    except Exception:
        toolbox_instances = set()
    # NOTE: API authorization is enforced by service, and we want different rules
    # for each toolbox, so we combine service and instance in this case to properly
    # allow that to happen.
    return list(list_services()) + sorted(
        f"{TOOLBOX_MOCK_SERVICE}-{instance}" for instance in toolbox_instances
    )


def _get_kubectl_wrapper(cluster: str) -> str:
    kubectl_wrapper = f"kubectl-eks-{cluster}"
    if not shutil.which(kubectl_wrapper):
        kubectl_wrapper = f"kubectl-{cluster}"
    return kubectl_wrapper


def parse_error(body: str) -> str:
    try:
        body_object = json.loads(body)
    except json.decoder.JSONDecodeError:
        return body
    return (
        body_object.get("reason")
        or body_object.get("message")
        or json.dumps(body_object, indent=4)
    )


def paasta_remote_run_copy(
    args: argparse.Namespace,
    system_paasta_config: SystemPaastaConfig,
) -> int:
    client = get_paasta_oapi_client_with_auth(
        cluster=get_paasta_oapi_api_clustername(cluster=args.cluster, is_eks=True),
        system_paasta_config=system_paasta_config,
    )
    if not client:
        print("Cannot get a paasta-api client")
        return 1

    # Create the config and extract the job name
    user = get_username()
    deployment_config = load_eks_or_adhoc_deployment_config(
        args.service, args.instance, args.cluster, args.toolbox, user
    )
    deployment_name = deployment_config.get_sanitised_deployment_name()
    job_name = format_remote_run_job_name(deployment_name, user)

    poll_response = client.remote_run.remote_run_poll(
        service=args.service,
        instance=args.instance,
        job_name=job_name,
        user=user,
        toolbox=args.toolbox,
    )
    if poll_response.status != 200:
        print(
            "Unable to find running remote-run pod: have you started one with `paasta remote-run start`?"
        )
        return 1

    if args.to_pod:
        # Create the toolbox copy command
        exec_command = f"scp -A {args.copy_file_source} {poll_response.pod_address}:{args.copy_file_dest}"
        # Pick the correct template and force /tmp/ for non-toolbox
        kubectl_cp_template = KUBECTL_CP_TO_CMD_TEMPLATE
        if not args.copy_file_dest.startswith("/tmp") and not args.toolbox:
            args.copy_file_dest = os.path.join("/tmp/", args.copy_file_dest)
    else:
        exec_command = f"scp -A {poll_response.pod_address}:{args.copy_file_source} {args.copy_file_dest}"
        kubectl_cp_template = KUBECTL_CP_FROM_CMD_TEMPLATE

    # Kubectl cp doesnt like directories as a destination
    if os.path.isdir(args.copy_file_dest):
        args.copy_file_dest = os.path.join(
            args.copy_file_dest, os.path.basename(args.copy_file_source)
        )

    if args.toolbox:
        cp_command = exec_command.split(" ")
    else:
        token_response = client.remote_run.remote_run_token(
            args.service, args.instance, user
        )
        kubectl_wrapper = _get_kubectl_wrapper(args.cluster)
        cp_command = kubectl_cp_template.format(
            kubectl_wrapper=kubectl_wrapper,
            namespace=poll_response.namespace,
            pod=poll_response.pod_name,
            source=args.copy_file_source,
            dest=args.copy_file_dest,
            token=token_response.token,
        ).split(" ")
    call = subprocess.run(cp_command, capture_output=True)
    if call.returncode != 0:
        print("Error copying file from remote-run pod: ", file=sys.stderr)
        print(call.stderr.decode("utf-8"), file=sys.stderr)
        return 1
    print(f"{args.copy_file_source} successfully copied to {args.copy_file_dest}")

    return 0


def paasta_remote_run_start(
    args: argparse.Namespace,
    system_paasta_config: SystemPaastaConfig,
    recursed: bool = False,
) -> int:
    status_prefix = "\x1b[2K\r"  # Clear line, carriage return
    client = get_paasta_oapi_client_with_auth(
        cluster=get_paasta_oapi_api_clustername(cluster=args.cluster, is_eks=True),
        system_paasta_config=system_paasta_config,
    )
    if not client:
        print("Cannot get a paasta-api client")
        return 1
    user = get_username()
    try:
        start_response = client.remote_run.remote_run_start(
            args.service,
            args.instance,
            RemoteRunStart(
                user=user,
                interactive=args.interactive,
                recreate=args.recreate,
                max_duration=args.max_duration,
                toolbox=args.toolbox,
                command=args.cmd or "",
            ),
        )
    except ApiException as e:
        error_msg = parse_error(e.body)
        print(f"Error from PaaSTA APIs while starting job: {error_msg}")
        return 1

    print(
        f"Triggered remote-run job for {args.service}. Waiting for pod to come online..."
    )
    start_time = time.time()
    while time.time() - start_time < args.timeout:
        poll_response = client.remote_run.remote_run_poll(
            service=args.service,
            instance=args.instance,
            job_name=start_response.job_name,
            user=user,
            toolbox=args.toolbox,
        )
        if poll_response.status == 200:
            print("")
            break
        print(f"{status_prefix}Status: {poll_response.message}", end="")
        if poll_response.status == 404:
            # Probably indicates a pod was terminating. Now that its gone, retry the whole process
            if not recursed:
                print("\nPod finished terminating. Rerunning")
                return paasta_remote_run_start(args, system_paasta_config, True)
            else:
                print("\nSomething went wrong. Pod still not found.")
                return 1
        time.sleep(10)
    else:
        print(f"{status_prefix}Timed out while waiting for job to start")
        return 1

    if not (args.interactive or args.toolbox or args.follow):
        print("Successfully started remote-run job")
        return 0

    print("Pod ready, establishing interactive session...")

    if args.toolbox:
        # NOTE: we only do this for toolbox containers since those images are built with interactive
        # access in mind, and SSH sessions provide better auditability of user actions.
        # I.e., being `nobody` is fine in a normal remote-run, but in toolbox containers
        # we will require knowing the real user (and some tools may need that too).
        exec_command = f"ssh -A {poll_response.pod_address}"
    else:
        token_response = client.remote_run.remote_run_token(
            args.service, args.instance, user
        )
        kubectl_wrapper = _get_kubectl_wrapper(args.cluster)
        exec_command = (
            KUBECTL_LOGS_CMD_TEMPLATE if args.follow else KUBECTL_EXEC_CMD_TEMPLATE
        ).format(
            kubectl_wrapper=kubectl_wrapper,
            namespace=poll_response.namespace,
            pod=poll_response.pod_name,
            token=token_response.token,
        )

    if args.copy_file:
        for filename in args.copy_file:
            cp_command = KUBECTL_CP_TO_CMD_TEMPLATE.format(
                kubectl_wrapper=kubectl_wrapper,
                namespace=poll_response.namespace,
                pod=poll_response.pod_name,
                source=filename,
                dest=os.path.join("/tmp", os.path.basename(filename)),
                token=token_response.token,
            ).split(" ")
            call = subprocess.run(cp_command, capture_output=True)
            if call.returncode != 0:
                print("Error copying file to remote-run pod: ", file=sys.stderr)
                print(call.stderr.decode("utf-8"), file=sys.stderr)
                return 1

    run_interactive_cli(exec_command)
    return 0


def paasta_remote_run_stop(
    args: argparse.Namespace,
    system_paasta_config: SystemPaastaConfig,
) -> int:
    client = get_paasta_oapi_client_with_auth(
        cluster=get_paasta_oapi_api_clustername(cluster=args.cluster, is_eks=True),
        system_paasta_config=system_paasta_config,
    )
    if not client:
        print("Cannot get a paasta-api client")
        return 1
    try:
        response = client.remote_run.remote_run_stop(
            args.service,
            args.instance,
            RemoteRunStop(user=get_username(), toolbox=args.toolbox),
        )
    except ApiException as e:
        error_msg = parse_error(e.body)
        print(f"Error from PaaSTA APIs while stopping job: {error_msg}")
        return 1
    print(response.message)
    return 0


def add_common_args_to_parser(parser: argparse.ArgumentParser):
    def _validate_service_name(name: str) -> str:
        # remove trailing slash if present for folks tab-completing directories
        if name.rstrip("/") not in _list_services_and_toolboxes():
            raise ValueError(f"{name} is not a known service name")
        return name

    service_arg = parser.add_argument(
        "-s",
        "--service",
        help="The name of the service you wish to inspect. Required.",
        required=True,
        # not using `choices` for validation to avoid the help text
        # for the command being incredibly large
        type=_validate_service_name,
    )
    service_arg.completer = lazy_choices_completer(_list_services_and_toolboxes)  # type: ignore
    instance_or_toolbox = parser.add_mutually_exclusive_group()
    instance_or_toolbox.add_argument(
        "-i",
        "--instance",
        help=(
            "Simulate a docker run for a particular instance of the "
            "service, like 'main' or 'canary'. Required."
        ),
        default="main",
    )
    instance_or_toolbox.add_argument(
        "--toolbox",
        help="The selected service is a 'toolbox' container",
        action="store_true",
        default=False,
    )
    cluster_arg = parser.add_argument(
        "-c",
        "--cluster",
        help="The name of the cluster you wish to run your task on. Required.",
        required=True,
        choices=list_clusters(),
    )
    cluster_arg.completer = lazy_choices_completer(list_clusters)  # type: ignore


def add_subparser(subparsers: argparse._SubParsersAction) -> None:
    remote_run_parser = subparsers.add_parser(
        "remote-run",
        help="Run services / jobs remotely",
        description="'paasta remote-run' runs services / jobs remotely",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers = remote_run_parser.add_subparsers(dest="remote_run_command")
    start_parser = subparsers.add_parser(
        "start",
        help="Start or connect to a remote-run job",
        description="Starts or connects to a remote-run-job",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    cmd_or_interactive = start_parser.add_mutually_exclusive_group()
    cmd_or_interactive.add_argument(
        "-I",
        "--interactive",
        help=(
            "Run container in interactive mode. If interactive is set the "
            'default command will be "bash" unless otherwise set by the "--cmd" flag'
        ),
        action="store_true",
        default=False,
    )
    cmd_or_interactive.add_argument(
        "-C",
        "--cmd",
        help=(
            "Run container with particular command, rather than the one specified in soa-configs"
        ),
        required=False,
        default=None,
    )
    start_parser.add_argument(
        "-m",
        "--max-duration",
        help=(
            "Amount of time in seconds after which the job is "
            "automatically stopped (capped by the API backend)"
        ),
        type=int,
        default=7200,
    )
    start_parser.add_argument(
        "-f",
        "--follow",
        help="Attach to and follow container output",
        action="store_true",
        default=False,
    )
    start_parser.add_argument(
        "-r",
        "--recreate",
        help="Recreate remote-run job if already existing",
        action="store_true",
        default=False,
    )
    start_parser.add_argument(
        "-t",
        "--timeout",
        help="Maximum time to wait for a job to start, in seconds",
        type=int,
        default=600,
    )
    start_parser.add_argument(
        "--copy-file",
        help="Adds a local file to /tmp inside the pod",
        type=str,
        action="append",
    )
    stop_parser = subparsers.add_parser(
        "stop",
        help="Stop your remote-run job if it exists",
        description="Stop your remote-run job if it exists",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    copy_parser = subparsers.add_parser(
        "copy",
        help="Copy a file from a running remote-run job",
        description="Copy a file from a running remote-run job",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    copy_parser.add_argument(
        "copy_file_source",
        help="Location of the file within the pod",
        type=str,
    )
    copy_parser.add_argument(
        "copy_file_dest",
        help="Location on the local host to copy the file to",
        type=str,
    )
    copy_parser.add_argument(
        "--to-pod",
        help="Copy a file to the pod instead of from the pod",
        action="store_true",
        default=False,
    )
    add_common_args_to_parser(start_parser)
    add_common_args_to_parser(stop_parser)
    add_common_args_to_parser(copy_parser)
    remote_run_parser.set_defaults(command=paasta_remote_run)


def paasta_remote_run(args: argparse.Namespace) -> int:
    system_paasta_config = load_system_paasta_config()
    if args.remote_run_command == "start":
        return paasta_remote_run_start(args, system_paasta_config)
    elif args.remote_run_command == "stop":
        return paasta_remote_run_stop(args, system_paasta_config)
    elif args.remote_run_command == "copy":
        return paasta_remote_run_copy(args, system_paasta_config)
    raise ValueError(f"Unsupported subcommand: {args.remote_run_command}")



# Copyright 2015-2018 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import mock

import pytest

from paasta_tools import setup_kubernetes_cr
from paasta_tools.kubernetes_tools import KubeCustomResource
from paasta_tools.utils import SystemPaastaConfig


def test_main():
    with mock.patch(
        "paasta_tools.setup_kubernetes_cr.KubeClient", autospec=True
    ), mock.patch(
        "paasta_tools.setup_kubernetes_cr.setup_all_custom_resources", autospec=True
    ) as mock_setup:
        mock_setup.return_value = True
        with pytest.raises(SystemExit) as e:
            setup_kubernetes_cr.main()
            assert e.value.code == 0

        mock_setup.return_value = False
        with pytest.raises(SystemExit) as e:
            setup_kubernetes_cr.main()
            assert e.value.code == 1


def test_setup_all_custom_resources():
    with mock.patch(
        "paasta_tools.setup_kubernetes_cr.ensure_namespace", autospec=True
    ), mock.patch(
        "paasta_tools.setup_kubernetes_cr.load_all_configs", autospec=True
    ), mock.patch(
        "paasta_tools.setup_kubernetes_cr.setup_custom_resources", autospec=True
    ) as mock_setup, mock.patch(
        "paasta_tools.setup_kubernetes_cr.load_custom_resource_definitions",
        autospec=True,
    ) as mock_load_custom_resources:
        mock_system_config = mock.Mock(
            get_cluster=mock.Mock(return_value="westeros-prod")
        )
        # if some CRs setup okay should return True
        mock_setup.side_effect = [True, False]

        mock_client = mock.Mock()
        flink_crd = mock.Mock()
        flink_crd.spec.names = mock.Mock(plural="flinkclusters", kind="FlinkCluster")
        cassandra_crd = mock.Mock()
        cassandra_crd.spec.names = mock.Mock(
            plural="cassandraclusters", kind="CassandraCluster"
        )
        mock_client.apiextensions_v1_beta1.list_custom_resource_definition.return_value = mock.Mock(
            items=[flink_crd, cassandra_crd]
        )

        mock_client.apiextensions.list_custom_resource_definition.return_value = (
            mock.Mock(items=[])
        )

        custom_resource_definitions = [
            mock.Mock(
                kube_kind=mock.Mock(plural="flinkclusters", singular="FlinkCluster")
            ),
            mock.Mock(
                kube_kind=mock.Mock(
                    plural="cassandraclusters", singular="CassandraCluster"
                )
            ),
        ]

        assert setup_kubernetes_cr.setup_all_custom_resources(
            mock_client,
            "/nail/soa",
            mock_system_config,
            custom_resource_definitions=custom_resource_definitions,
        )

        mock_load_custom_resources.return_value = [
            mock.Mock(plural="flinks"),
            mock.Mock(plural="cassandraclusters"),
        ]
        mock_setup.side_effect = [True, True]
        mock_system_config = mock.Mock(
            get_cluster=mock.Mock(return_value="westeros-prod")
        )
        # if all CRs setup okay should return True
        assert setup_kubernetes_cr.setup_all_custom_resources(
            mock_client,
            "/nail/soa",
            mock_system_config,
            custom_resource_definitions=custom_resource_definitions,
        )

        mock_load_custom_resources.return_value = []
        mock_system_config = mock.Mock(
            get_cluster=mock.Mock(return_value="westeros-prod")
        )
        assert setup_kubernetes_cr.setup_all_custom_resources(
            mock_client, "/nail/soa", mock_system_config, custom_resource_definitions=[]
        )

        mock_setup.side_effect = []
        # if no CRs setup should return True
        assert setup_kubernetes_cr.setup_all_custom_resources(
            mock_client, "/nail/soa", mock_system_config, custom_resource_definitions=[]
        )

        mock_setup.side_effect = [False, False]
        # if all CRs setup fail should return False
        assert setup_kubernetes_cr.setup_all_custom_resources(
            mock_client, "/nail/soa", mock_system_config, custom_resource_definitions=[]
        )


def test_setup_all_custom_resources_flink():
    with mock.patch(
        "paasta_tools.setup_kubernetes_cr.ensure_namespace", autospec=True
    ), mock.patch(
        "paasta_tools.setup_kubernetes_cr.load_all_configs", autospec=True
    ) as mock_load_all, mock.patch(
        "paasta_tools.setup_kubernetes_cr.setup_custom_resources", autospec=True
    ) as mock_setup:
        mock_system_config = mock.Mock(
            get_cluster=mock.Mock(return_value="westeros-prod")
        )
        # if some CRs setup okay should return True
        mock_setup.side_effect = [True, False]

        mock_client = mock.Mock()
        flink_crd = mock.Mock()
        flink_crd.spec.names = mock.Mock(
            singular="flink", plural="flinks", kind="flink"
        )

        mock_client.apiextensions.list_custom_resource_definition.return_value = (
            mock.Mock(items=[flink_crd])
        )

        mock_client.apiextensions_v1_beta1.list_custom_resource_definition.return_value = mock.Mock(
            items=[]
        )

        custom_resource_definitions = [
            mock.Mock(
                kube_kind=mock.Mock(plural="flinks", singular="flink", kind="Flink")
            ),
        ]

        setup_kubernetes_cr.setup_all_custom_resources(
            mock_client,
            "/nail/soa",
            mock_system_config,
            custom_resource_definitions=custom_resource_definitions,
        )

        assert mock_load_all.called


def test_load_all_configs():
    with mock.patch(
        "paasta_tools.utils.load_service_instance_configs",
        autospec=True,
    ) as mock_load_configs, mock.patch("os.listdir", autospec=True) as mock_oslist:
        mock_oslist.return_value = ["kurupt", "mc"]
        ret = setup_kubernetes_cr.load_all_configs(
            cluster="westeros-prod", file_prefix="thing", soa_dir="/nail/soa"
        )
        mock_load_configs.assert_has_calls(
            [
                mock.call("mc", "thing", "westeros-prod", soa_dir="/nail/soa"),
                mock.call("kurupt", "thing", "westeros-prod", soa_dir="/nail/soa"),
            ],
            any_order=True,
        )
        assert "kurupt" in ret.keys()
        assert "mc" in ret.keys()


def test_load_all_flink_configs():
    with mock.patch(
        "paasta_tools.utils.load_service_instance_configs",
        autospec=True,
    ) as mock_load_configs, mock.patch("os.listdir", autospec=True) as mock_oslist:
        mock_oslist.return_value = ["kurupt", "mc"]
        mock_load_configs.side_effect = [
            {
                "foo": {"mem": 2},
                "bar": {"cpus": 3},
            },
            {
                "bar": {"cpus": 3},
            },
        ]
        ret = setup_kubernetes_cr.load_all_configs(
            cluster="westeros-prod", file_prefix="flink", soa_dir="/nail/soa"
        )

        mock_load_configs.assert_has_calls(
            [
                mock.call("kurupt", "flink", "westeros-prod", soa_dir="/nail/soa"),
                mock.call("mc", "flink", "westeros-prod", soa_dir="/nail/soa"),
            ],
            any_order=True,
        )

        assert "kurupt" in ret.keys()
        assert "mc" in ret.keys()

        assert ret["kurupt"] == {"foo": {"mem": 2}, "bar": {"cpus": 3}}
        assert ret["mc"] == {"bar": {"cpus": 3}}


def test_setup_custom_resources():
    with mock.patch(
        "paasta_tools.setup_kubernetes_cr.list_custom_resources", autospec=True
    ) as mock_list_cr, mock.patch(
        "paasta_tools.setup_kubernetes_cr.reconcile_kubernetes_resource", autospec=True
    ) as mock_reconcile_kubernetes_resource:
        mock_client = mock.Mock()
        mock_kind = mock.Mock()
        mock_crd = mock.Mock()
        assert setup_kubernetes_cr.setup_custom_resources(
            kube_client=mock_client,
            kind=mock_kind,
            version="v1",
            config_dicts={},
            group="yelp.com",
            cluster="mycluster",
            crd=mock_crd,
        )

        mock_reconcile_kubernetes_resource.side_effect = [True, False]
        assert not setup_kubernetes_cr.setup_custom_resources(
            kube_client=mock_client,
            kind=mock_kind,
            version="v1",
            config_dicts={"kurupt": "something", "mc": "another"},
            group="yelp.com",
            cluster="mycluster",
            crd=mock_crd,
        )

        mock_reconcile_kubernetes_resource.side_effect = [True, True]
        assert setup_kubernetes_cr.setup_custom_resources(
            kube_client=mock_client,
            kind=mock_kind,
            version="v1",
            config_dicts={"kurupt": "something", "mc": "another"},
            group="yelp.com",
            cluster="mycluster",
            crd=mock_crd,
        )
        mock_reconcile_kubernetes_resource.assert_has_calls(
            [
                mock.call(
                    kube_client=mock_client,
                    service="kurupt",
                    instance_configs="something",
                    cluster="mycluster",
                    instance=None,
                    kind=mock_kind,
                    custom_resources=mock_list_cr.return_value,
                    version="v1",
                    group="yelp.com",
                    crd=mock_crd,
                ),
                mock.call(
                    kube_client=mock_client,
                    service="mc",
                    instance_configs="another",
                    cluster="mycluster",
                    instance=None,
                    kind=mock_kind,
                    custom_resources=mock_list_cr.return_value,
                    version="v1",
                    group="yelp.com",
                    crd=mock_crd,
                ),
            ]
        )


def test_format_custom_resource():
    with mock.patch(
        "paasta_tools.setup_kubernetes_cr.get_config_hash", autospec=True
    ) as mock_get_config_hash, mock.patch(
        "paasta_tools.setup_kubernetes_cr.load_system_paasta_config", autospec=True
    ) as mock_load_system_paasta_config:
        mock_load_system_paasta_config.return_value = SystemPaastaConfig(
            {"dashboard_links": {}, "cr_owners": {"flink": "stream-processing"}}, ""
        )
        expected = {
            "apiVersion": "yelp.com/v1",
            "kind": "flink",
            "metadata": {
                "name": "kurupt--fm-radio--station",
                "namespace": "paasta-flinks",
                "labels": {
                    "yelp.com/paasta_service": "kurupt_fm",
                    "yelp.com/paasta_instance": "radio_station",
                    "yelp.com/paasta_cluster": "mycluster",
                    "yelp.com/owner": "stream-processing",
                    "yelp.com/paasta_config_sha": mock_get_config_hash.return_value,
                    "paasta.yelp.com/service": "kurupt_fm",
                    "paasta.yelp.com/instance": "radio_station",
                    "paasta.yelp.com/cluster": "mycluster",
                    "paasta.yelp.com/config_sha": mock_get_config_hash.return_value,
                    "paasta.yelp.com/git_sha": "gitsha",
                },
                "annotations": {
                    "yelp.com/desired_state": "running",
                    "paasta.yelp.com/desired_state": "running",
                    "paasta.yelp.com/dashboard_base_url": "http://flink.k8s.mycluster.paasta:31080/",
                },
            },
            "spec": {"dummy": "conf"},
        }
        assert (
            setup_kubernetes_cr.format_custom_resource(
                instance_config={"dummy": "conf"},
                service="kurupt_fm",
                instance="radio_station",
                cluster="mycluster",
                kind="flink",
                version="v1",
                group="yelp.com",
                namespace="paasta-flinks",
                git_sha="gitsha",
                is_eks=False,
            )
            == expected
        )


def test_paasta_config_flink_dashboard_base_url():
    with mock.patch(
        "paasta_tools.setup_kubernetes_cr.load_system_paasta_config", autospec=True
    ) as mock_load_system_paasta_config:
        mock_load_system_paasta_config.return_value = SystemPaastaConfig(
            {
                "dashboard_links": {
                    "mycluster": {"Flink": "http://flink.mycluster.paasta"}
                }
            },
            "",
        )
        expected = "http://flink.mycluster.paasta/"
        assert (
            setup_kubernetes_cr.get_dashboard_base_url(
                kind="flink", cluster="mycluster", is_eks=False
            )
            == expected
        )


@mock.patch(
    "paasta_tools.setup_kubernetes_cr.LONG_RUNNING_INSTANCE_TYPE_HANDLERS",
    autospec=True,
)
def test_reconcile_kubernetes_resource(mock_LONG_RUNNING_INSTANCE_TYPE_HANDLERS):
    with mock.patch(
        "paasta_tools.setup_kubernetes_cr.format_custom_resource", autospec=True
    ) as mock_format_custom_resource, mock.patch(
        "paasta_tools.setup_kubernetes_cr.create_custom_resource", autospec=True
    ) as mock_create_custom_resource, mock.patch(
        "paasta_tools.setup_kubernetes_cr.update_custom_resource", autospec=True
    ) as mock_update_custom_resource:
        mock_kind = mock.Mock(singular="flink", plural="flinks")
        mock_custom_resources = [
            KubeCustomResource(
                service="kurupt",
                instance="fm",
                config_sha="conf123",
                git_sha="git123",
                kind="flink",
                name="foo",
                namespace="paasta-flinks",
            )
        ]
        mock_client = mock.Mock()
        # no instances, do nothing
        assert setup_kubernetes_cr.reconcile_kubernetes_resource(
            kube_client=mock_client,
            service="mc",
            instance_configs={},
            cluster="mycluster",
            custom_resources=mock_custom_resources,
            kind=mock_kind,
            version="v1",
            group="yelp.com",
            crd=mock.Mock(),
        )
        assert not mock_create_custom_resource.called
        assert not mock_update_custom_resource.called

        # instance up to date, do nothing
        mock_format_custom_resource.return_value = {
            "metadata": {
                "labels": {
                    "yelp.com/paasta_config_sha": "conf123",
                    "paasta.yelp.com/config_sha": "conf123",
                    "paasta.yelp.com/git_sha": "git123",
                },
                "name": "foo",
                "namespace": "paasta-flinks",
            }
        }
        assert setup_kubernetes_cr.reconcile_kubernetes_resource(
            kube_client=mock_client,
            service="kurupt",
            instance_configs={"fm": {"some": "config"}},
            cluster="cluster",
            custom_resources=mock_custom_resources,
            kind=mock_kind,
            version="v1",
            group="yelp.com",
            crd=mock.Mock(),
        )
        assert not mock_create_custom_resource.called
        assert not mock_update_custom_resource.called

        # instance diff config, update
        mock_format_custom_resource.return_value = {
            "metadata": {
                "labels": {"paasta.yelp.com/config_sha": "conf456"},
                "name": "foo",
                "namespace": "paasta-flinks",
            }
        }
        assert setup_kubernetes_cr.reconcile_kubernetes_resource(
            kube_client=mock_client,
            service="kurupt",
            instance_configs={"fm": {"some": "config"}},
            cluster="mycluster",
            custom_resources=mock_custom_resources,
            kind=mock_kind,
            version="v1",
            group="yelp.com",
            crd=mock.Mock(),
        )
        assert not mock_create_custom_resource.called
        mock_update_custom_resource.assert_called_with(
            kube_client=mock_client,
            name="kurupt-fm",
            version="v1",
            kind=mock_kind,
            formatted_resource=mock_format_custom_resource.return_value,
            group="yelp.com",
        )

        # instance not exist, create
        assert setup_kubernetes_cr.reconcile_kubernetes_resource(
            kube_client=mock_client,
            service="mc",
            instance_configs={"grindah": {"some": "conf"}},
            cluster="mycluster",
            custom_resources=mock_custom_resources,
            kind=mock_kind,
            version="v1",
            group="yelp.com",
            crd=mock.Mock(),
        )
        mock_create_custom_resource.assert_called_with(
            kube_client=mock_client,
            version="v1",
            kind=mock_kind,
            formatted_resource=mock_format_custom_resource.return_value,
            group="yelp.com",
        )

        # instance not exist, create but error with k8s
        mock_create_custom_resource.side_effect = Exception
        assert not setup_kubernetes_cr.reconcile_kubernetes_resource(
            kube_client=mock_client,
            service="mc",
            instance_configs={"grindah": {"some": "conf"}},
            cluster="mycluster",
            custom_resources=mock_custom_resources,
            kind=mock_kind,
            version="v1",
            group="yelp.com",
            crd=mock.Mock(),
        )
        mock_create_custom_resource.assert_called_with(
            kube_client=mock_client,
            version="v1",
            kind=mock_kind,
            formatted_resource=mock_format_custom_resource.return_value,
            group="yelp.com",
        )



from unittest import mock

from paasta_tools.flinkeks_tools import load_flinkeks_instance_config


def test_load_flinkeks_instance_config():
    with mock.patch(
        "paasta_tools.flinkeks_tools.load_v2_deployments_json", autospec=True
    ) as mock_load_v2_deployments_json, mock.patch(
        "paasta_tools.flinkeks_tools.load_service_instance_config", autospec=True
    ), mock.patch(
        "paasta_tools.flinkeks_tools.FlinkEksDeploymentConfig", autospec=True
    ) as mock_flinkeks_deployment_config:
        mock_config = {
            "port": None,
            "monitoring": {},
            "deploy": {},
            "data": {},
            "smartstack": {},
            "dependencies": {},
        }
        flinkeks_deployment_config = load_flinkeks_instance_config(
            service="fake_flinkservice",
            instance="fake_instance",
            cluster="fake_cluster",
            load_deployments=True,
            soa_dir="/foo/bar",
        )
        mock_load_v2_deployments_json.assert_called_with(
            service="fake_flinkservice", soa_dir="/foo/bar"
        )
        mock_flinkeks_deployment_config.assert_called_with(
            service="fake_flinkservice",
            instance="fake_instance",
            cluster="fake_cluster",
            config_dict=mock_config,
            branch_dict=mock_load_v2_deployments_json.return_value.get_branch_dict(),
            soa_dir="/foo/bar",
        )

        assert (
            flinkeks_deployment_config == mock_flinkeks_deployment_config.return_value
        )



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
PaaSTA service instance status/start/stop etc.
"""
import asyncio
import logging
import re
import traceback
from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional

import a_sync
from pyramid.request import Request
from pyramid.response import Response
from pyramid.view import view_config

import paasta_tools.mesos.exceptions as mesos_exceptions
from paasta_tools import tron_tools
from paasta_tools.api import settings
from paasta_tools.api.views.exception import ApiFailure
from paasta_tools.cli.cmds.status import get_actual_deployments
from paasta_tools.instance import kubernetes as pik
from paasta_tools.mesos_tools import get_all_frameworks as get_all_mesos_frameworks
from paasta_tools.utils import compose_job_id
from paasta_tools.utils import DeploymentVersion
from paasta_tools.utils import NoConfigurationForServiceError
from paasta_tools.utils import PAASTA_K8S_INSTANCE_TYPES
from paasta_tools.utils import TimeoutError
from paasta_tools.utils import validate_service_instance

log = logging.getLogger(__name__)


def tron_instance_status(
    instance_status: Mapping[str, Any], service: str, instance: str, verbose: int
) -> Mapping[str, Any]:
    status: Dict[str, Any] = {}
    client = tron_tools.get_tron_client()
    short_job, action = instance.split(".")
    job = f"{service}.{short_job}"
    job_content = client.get_job_content(job=job)

    try:
        latest_run_id = client.get_latest_job_run_id(job_content=job_content)
        if latest_run_id is None:
            action_run = {"state": "Hasn't run yet (no job run id found)"}
        else:
            action_run = client.get_action_run(
                job=job, action=action, run_id=latest_run_id
            )
    except Exception as e:
        action_run = {"state": f"Failed to get latest run info: {e}"}

    # job info
    status["job_name"] = short_job
    status["job_status"] = job_content["status"]
    status["job_schedule"] = "{} {}".format(
        job_content["scheduler"]["type"], job_content["scheduler"]["value"]
    )
    status["job_url"] = (
        tron_tools.get_tron_dashboard_for_cluster(settings.cluster) + f"#job/{job}"
    )

    if action:
        status["action_name"] = action
    if action_run.get("state"):
        status["action_state"] = action_run["state"]
    if action_run.get("start_time"):
        status["action_start_time"] = action_run["start_time"]
    if action_run.get("raw_command"):
        status["action_raw_command"] = action_run["raw_command"]
    if action_run.get("stdout"):
        status["action_stdout"] = "\n".join(action_run["stdout"])
    if action_run.get("stderr"):
        status["action_stderr"] = "\n".join(action_run["stderr"])
    if action_run.get("command"):
        status["action_command"] = action_run["command"]

    return status


def legacy_remote_run_filter_frameworks(service, instance, frameworks=None):
    if frameworks is None:
        frameworks = get_all_mesos_frameworks(active_only=True)

    prefix = f"paasta-remote {service}.{instance}"
    return [f for f in frameworks if f.name.startswith(prefix)]


def adhoc_instance_status(
    instance_status: Mapping[str, Any], service: str, instance: str, verbose: int
) -> List[Dict[str, Any]]:
    status = []
    filtered = legacy_remote_run_filter_frameworks(service, instance)
    filtered.sort(key=lambda x: x.name)
    for f in filtered:
        launch_time, run_id = re.match(
            r"paasta-remote [^\s]+ (\w+) (\w+)", f.name
        ).groups()
        status.append(
            {"launch_time": launch_time, "run_id": run_id, "framework_id": f.id}
        )
    return status


async def _task_result_or_error(future):
    try:
        return {"value": await future}
    except (AttributeError, mesos_exceptions.SlaveDoesNotExist):
        return {"error_message": "None"}
    except TimeoutError:
        return {"error_message": "Timed Out"}
    except Exception:
        return {"error_message": "Unknown"}


def no_configuration_for_service_message(cluster, service, instance):
    return (
        f"No instance named '{compose_job_id(service, instance)}' has been "
        f"configured to run in the {settings.cluster} cluster"
    )


@view_config(
    route_name="service.instance.status", request_method="GET", renderer="json"
)
def instance_status(
    request: Request,
) -> dict[str, Any]:  # godspeed to anyone typing the retval here
    # NOTE: swagger_data is populated by pyramid_swagger
    service = request.swagger_data.get("service")
    instance = request.swagger_data.get("instance")
    verbose = request.swagger_data.get("verbose") or 0
    use_new = request.swagger_data.get("new") or False
    all_namespaces = request.swagger_data.get("all_namespaces") or False
    include_envoy = request.swagger_data.get("include_envoy")
    if include_envoy is None:
        include_envoy = True
    include_mesos = request.swagger_data.get("include_mesos")
    if include_mesos is None:
        include_mesos = True

    instance_status: Dict[str, Any] = {}
    instance_status["service"] = service
    instance_status["instance"] = instance
    try:
        instance_type = validate_service_instance(
            service, instance, settings.cluster, settings.soa_dir
        )
    except NoConfigurationForServiceError:
        error_message = no_configuration_for_service_message(
            settings.cluster,
            service,
            instance,
        )
        raise ApiFailure(error_message, 404)
    except Exception:
        error_message = traceback.format_exc()
        raise ApiFailure(error_message, 500)

    if instance_type != "tron":
        try:
            actual_deployments = get_actual_deployments(service, settings.soa_dir)
        except Exception:
            error_message = traceback.format_exc()
            raise ApiFailure(error_message, 500)

        version = get_deployment_version(actual_deployments, settings.cluster, instance)
        # exit if the deployment key is not found
        if not version:
            error_message = (
                "Deployment key %s not found.  Try to execute the corresponding pipeline if it's a fresh instance"
                % ".".join([settings.cluster, instance])
            )
            raise ApiFailure(error_message, 404)

        instance_status["version"] = version.short_sha_repr()
        # Kept for backwards compatibility
        # TODO: Remove once all clients+clusters updated to use deploymentversion
        instance_status["git_sha"] = version.sha[:8]
    else:
        instance_status["version"] = ""
        instance_status["git_sha"] = ""
    try:
        if instance_type == "adhoc":
            instance_status["adhoc"] = adhoc_instance_status(
                instance_status, service, instance, verbose
            )
        elif pik.can_handle(instance_type):
            instance_status.update(
                pik.instance_status(
                    service=service,
                    instance=instance,
                    verbose=verbose,
                    include_envoy=include_envoy,
                    use_new=use_new,
                    instance_type=instance_type,
                    settings=settings,
                    all_namespaces=all_namespaces,
                )
            )
        elif instance_type == "tron":
            instance_status["tron"] = tron_instance_status(
                instance_status, service, instance, verbose
            )
        else:
            error_message = (
                f"Unknown instance_type {instance_type} of {service}.{instance}"
            )
            raise ApiFailure(error_message, 404)
    except Exception:
        error_message = traceback.format_exc()
        raise ApiFailure(error_message, 500)

    return instance_status


@view_config(
    route_name="service.instance.set_state", request_method="POST", renderer="json"
)
def instance_set_state(
    request,
) -> None:
    service = request.swagger_data.get("service")
    instance = request.swagger_data.get("instance")
    desired_state = request.swagger_data.get("desired_state")

    try:
        instance_type = validate_service_instance(
            service, instance, settings.cluster, settings.soa_dir
        )
    except NoConfigurationForServiceError:
        error_message = no_configuration_for_service_message(
            settings.cluster,
            service,
            instance,
        )
        raise ApiFailure(error_message, 404)
    except Exception:
        error_message = traceback.format_exc()
        raise ApiFailure(error_message, 500)

    if pik.can_set_state(instance_type):
        try:
            pik.set_cr_desired_state(
                kube_client=settings.kubernetes_client,
                service=service,
                instance=instance,
                instance_type=instance_type,
                desired_state=desired_state,
            )
        except RuntimeError as e:
            raise ApiFailure(e, 500)
    else:
        error_message = (
            f"instance_type {instance_type} of {service}.{instance} doesn't "
            "support set_state"
        )
        raise ApiFailure(error_message, 500)


@view_config(
    route_name="service.instance.bounce_status",
    request_method="GET",
    renderer="json",
)
def bounce_status(request):
    service = request.swagger_data.get("service")
    instance = request.swagger_data.get("instance")
    try:
        instance_type = validate_service_instance(
            service, instance, settings.cluster, settings.soa_dir
        )
    except NoConfigurationForServiceError:
        error_message = no_configuration_for_service_message(
            settings.cluster,
            service,
            instance,
        )
        raise ApiFailure(error_message, 404)
    except Exception:
        error_message = traceback.format_exc()
        raise ApiFailure(error_message, 500)

    if instance_type not in PAASTA_K8S_INSTANCE_TYPES:
        # We are using HTTP 204 to indicate that the instance exists but has
        # no bounce status to be returned.  The client should just mark the
        # instance as bounced.
        response = Response()
        response.status_int = 204
        return response

    try:
        return pik.bounce_status(
            service, instance, settings, is_eks=(instance_type == "eks")
        )
    except NoConfigurationForServiceError:
        # Handle race condition where instance has been removed since the above validation
        error_message = no_configuration_for_service_message(
            settings.cluster,
            service,
            instance,
        )
        raise ApiFailure(error_message, 404)
    except asyncio.TimeoutError:
        raise ApiFailure(
            "Temporary issue fetching bounce status. Please try again.", 599
        )
    except Exception as e:
        error_message = traceback.format_exc()
        if getattr(e, "status", None) == 404:
            # some bounces delete the app & recreate
            # in this case, we relay the 404 and cli handles gracefully
            raise ApiFailure(error_message, 404)
        # for all others, treat as a 500
        raise ApiFailure(error_message, 500)


def add_executor_info(task):
    task._Task__items["executor"] = a_sync.block(task.executor).copy()
    task._Task__items["executor"].pop("tasks", None)
    task._Task__items["executor"].pop("completed_tasks", None)
    task._Task__items["executor"].pop("queued_tasks", None)
    return task


def add_slave_info(task):
    task._Task__items["slave"] = a_sync.block(task.slave)._MesosSlave__items.copy()
    return task


def get_deployment_version(
    actual_deployments: Mapping[str, DeploymentVersion], cluster: str, instance: str
) -> Optional[DeploymentVersion]:
    key = ".".join((cluster, instance))
    return actual_deployments[key] if key in actual_deployments else None


@view_config(
    route_name="service.instance.mesh_status",
    request_method="GET",
    renderer="json",
)
def instance_mesh_status(
    request: Request,
) -> dict[str, Any]:  # godspeed to anyone typing the retval here
    # NOTE: swagger_data is populated by pyramid_swagger
    service = request.swagger_data.get("service")
    instance = request.swagger_data.get("instance")
    include_envoy = request.swagger_data.get("include_envoy")

    instance_mesh: Dict[str, Any] = {}
    instance_mesh["service"] = service
    instance_mesh["instance"] = instance

    try:
        instance_type = validate_service_instance(
            service, instance, settings.cluster, settings.soa_dir
        )
    except NoConfigurationForServiceError:
        error_message = (
            f"No instance named '{compose_job_id(service, instance)}' has been "
            f"configured to run in the {settings.cluster} cluster"
        )
        raise ApiFailure(error_message, 404)
    except Exception:
        error_message = traceback.format_exc()
        raise ApiFailure(error_message, 500)

    try:
        instance_mesh.update(
            pik.kubernetes_mesh_status(
                service=service,
                instance=instance,
                instance_type=instance_type,
                settings=settings,
                include_envoy=include_envoy,
            )
        )
    except RuntimeError as e:
        raise ApiFailure(str(e), 405)
    except Exception:
        error_message = traceback.format_exc()
        raise ApiFailure(error_message, 500)

    return instance_mesh



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import mock

from paasta_tools import generate_deployments_for_service
from paasta_tools.long_running_service_tools import LongRunningServiceConfig


def test_get_deploy_group_mappings():
    fake_service = "fake_service"
    fake_soa_dir = "/no/yes/maybe"

    fake_service_configs = [
        LongRunningServiceConfig(
            service=fake_service,
            cluster="clusterA",
            instance="main",
            branch_dict=None,
            config_dict={"deploy_group": "no_thanks"},
        ),
        LongRunningServiceConfig(
            service=fake_service,
            cluster="clusterB",
            instance="main",
            branch_dict=None,
            config_dict={"deploy_group": "try_me"},
        ),
        LongRunningServiceConfig(
            service=fake_service,
            cluster="clusterC",
            instance="main",
            branch_dict=None,
            config_dict={"deploy_group": "but-why"},
        ),
    ]

    fake_remote_refs = {
        "refs/tags/paasta-try_me-20160308T053933-deploy": "123456",
        "refs/tags/paasta-clusterB.main-123-stop": "123456",
        "refs/tags/paasta-okay-20160308T053933-deploy": "ijowarg",
        "refs/tags/paasta-no_thanks-20160308T053933-deploy": "789009",
        "refs/tags/paasta-nah-20160308T053933-deploy": "j8yiomwer",
        "refs/tags/paasta-but-why+extrastuff-20220308T053933-deploy": "123456",
    }

    expected = {
        "fake_service:paasta-clusterA.main": {
            "docker_image": "services-fake_service:paasta-789009",
            "desired_state": "start",
            "force_bounce": None,
        },
        "fake_service:paasta-clusterB.main": {
            "docker_image": "services-fake_service:paasta-123456",
            "desired_state": "stop",
            "force_bounce": "123",
        },
        "fake_service:paasta-clusterC.main": {
            "docker_image": "services-fake_service:paasta-123456-extrastuff",
            "desired_state": "start",
            "force_bounce": None,
        },
    }
    expected_v2 = {
        "deployments": {
            "try_me": {
                "docker_image": "services-fake_service:paasta-123456",
                "git_sha": "123456",
                "image_version": None,
            },
            "no_thanks": {
                "docker_image": "services-fake_service:paasta-789009",
                "git_sha": "789009",
                "image_version": None,
            },
            "but-why": {
                "docker_image": "services-fake_service:paasta-123456-extrastuff",
                "git_sha": "123456",
                "image_version": "extrastuff",
            },
        },
        "controls": {
            "fake_service:clusterA.main": {
                "desired_state": "start",
                "force_bounce": None,
            },
            "fake_service:clusterB.main": {
                "desired_state": "stop",
                "force_bounce": "123",
            },
            "fake_service:clusterC.main": {
                "desired_state": "start",
                "force_bounce": None,
            },
        },
    }
    with mock.patch(
        "paasta_tools.generate_deployments_for_service.get_instance_configs_for_service",
        return_value=fake_service_configs,
        autospec=True,
    ) as get_instance_configs_for_service_patch, mock.patch(
        "paasta_tools.remote_git.list_remote_refs",
        return_value=fake_remote_refs,
        autospec=True,
    ) as list_remote_refs_patch:
        actual, actual_v2 = generate_deployments_for_service.get_deploy_group_mappings(
            fake_soa_dir, fake_service
        )
        get_instance_configs_for_service_patch.assert_called_once_with(
            soa_dir=fake_soa_dir, service=fake_service
        )
        assert list_remote_refs_patch.call_count == 1
        assert expected == actual
        assert expected_v2 == actual_v2


def test_main():
    fake_soa_dir = "/etc/true/null"
    file_mock = mock.mock_open()
    with mock.patch(
        "paasta_tools.generate_deployments_for_service.parse_args",
        return_value=mock.Mock(
            verbose=False, soa_dir=fake_soa_dir, service="fake_service"
        ),
        autospec=True,
    ) as parse_patch, mock.patch(
        "os.path.abspath", return_value="ABSOLUTE", autospec=True
    ) as abspath_patch, mock.patch(
        "paasta_tools.generate_deployments_for_service.get_deploy_group_mappings",
        return_value=(
            {"MAP": {"docker_image": "PINGS", "desired_state": "start"}},
            mock.sentinel.v2_mappings,
        ),
        autospec=True,
    ) as mappings_patch, mock.patch(
        "os.path.join", return_value="JOIN", autospec=True
    ) as join_patch, mock.patch(
        "builtins.open", file_mock, autospec=None
    ) as open_patch, mock.patch(
        "json.dump", autospec=True
    ) as json_dump_patch, mock.patch(
        "json.load", return_value={"OLD_MAP": "PINGS"}, autospec=True
    ) as json_load_patch, mock.patch(
        "paasta_tools.generate_deployments_for_service.atomic_file_write", autospec=True
    ) as atomic_file_write_patch:
        generate_deployments_for_service.main()
        parse_patch.assert_called_once_with()
        abspath_patch.assert_called_once_with(fake_soa_dir)
        mappings_patch.assert_called_once_with(
            soa_dir="ABSOLUTE", service="fake_service"
        ),

        join_patch.assert_any_call(
            "ABSOLUTE", "fake_service", generate_deployments_for_service.TARGET_FILE
        ),
        assert join_patch.call_count == 2

        atomic_file_write_patch.assert_called_once_with("JOIN")
        open_patch.assert_called_once_with("JOIN", "r")
        json_dump_patch.assert_called_once_with(
            {
                "v1": {"MAP": {"docker_image": "PINGS", "desired_state": "start"}},
                "v2": mock.sentinel.v2_mappings,
            },
            atomic_file_write_patch.return_value.__enter__.return_value,
        )
        json_load_patch.assert_called_once_with(
            file_mock.return_value.__enter__.return_value
        )

        # test no update to file if content unchanged
        json_load_patch.return_value = {
            "v1": {"MAP": {"docker_image": "PINGS", "desired_state": "start"}},
            "v2": mock.sentinel.v2_mappings,
        }
        json_dump_patch.reset_mock()
        generate_deployments_for_service.main()
        assert not json_dump_patch.called

        # test IOError path
        open_patch.side_effect = IOError
        generate_deployments_for_service.main()
        assert json_dump_patch.called


def test_get_deployments_dict():
    branch_mappings = {
        "app1": {
            "docker_image": "image1",
            "desired_state": "start",
            "force_bounce": "1418951213",
        },
        "app2": {
            "docker_image": "image2",
            "desired_state": "stop",
            "force_bounce": "1412345678",
        },
    }

    v2_mappings = mock.sentinel.v2_mappings

    assert generate_deployments_for_service.get_deployments_dict_from_deploy_group_mappings(
        branch_mappings, v2_mappings
    ) == {
        "v1": branch_mappings,
        "v2": mock.sentinel.v2_mappings,
    }


def test_get_desired_state_understands_tags():
    remote_refs = {
        "refs/heads/master": "7894E99E6805E9DC8C1D8EB26229E3E2243878C9",
        "refs/remotes/origin/HEAD": "EE8796C4E4295B7D4087E3EB73662B99218DAD94",
        "refs/remotes/origin/master": "5F7C10B320A4EDBC4773C5FEFB1CD7B7A84FCB69",
        "refs/tags/paasta-paasta-cluster.instance-20150721T183905-start": "4EF01B5A574B519AB546309E89F72972A33B6B75",
        "refs/tags/paasta-paasta-cluster.instance-20151106T233211-stop": "A5AB2A012DC238D4F6DD269C40A4BD3A99D52B1F",
        "refs/tags/paasta-cluster.instance-20160202T233805-start": "BE68473F98F619F26FD7824B8F56F9A7ABAEB860",
        "refs/tags/paasta-cluster2.someinstance-20160202T233805-start": "D6B9A0F86DC54A132FBB7747460F53F48C9AEEAD",
        "refs/tags/paasta-cluster2.someinstance-20160205T182601-stop": "9085FD67ED1BB5FADAFA7F2AFAF8DEDEE7342711",
        "refs/tags/paasta-cluster.instance-20160308T053933-deploy": "4EF01B5A574B519AB546309E89F72972A33B6B75",
        "refs/tags/paasta-cluster2.someinstance-20160308T053933-deploy": "9085FD67ED1BB5FADAFA7F2AFAF8DEDEE7342711",
    }
    branch = "cluster2.someinstance"
    sha = "9085FD67ED1BB5FADAFA7F2AFAF8DEDEE7342711"
    expected_desired_state = ("stop", "20160205T182601")
    actual = generate_deployments_for_service.get_desired_state_by_branch_and_sha(
        remote_refs
    )[(branch, sha)]

    assert actual == expected_desired_state



# Copyright 2015-2019 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import argparse
import logging
import sys
from multiprocessing import Pool
from os import cpu_count
from typing import Any
from typing import Callable
from typing import cast
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Type

from mypy_extensions import Arg
from mypy_extensions import NamedArg

from paasta_tools.kubernetes_tools import get_all_managed_namespaces
from paasta_tools.kubernetes_tools import get_all_nodes
from paasta_tools.kubernetes_tools import get_all_pods
from paasta_tools.kubernetes_tools import group_pods_by_service_instance
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import V1Node
from paasta_tools.kubernetes_tools import V1Pod
from paasta_tools.metrics import metrics_lib
from paasta_tools.monitoring_tools import ReplicationChecker
from paasta_tools.paasta_service_config_loader import PaastaServiceConfigLoader
from paasta_tools.smartstack_tools import KubeSmartstackEnvoyReplicationChecker
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import InstanceConfig_T
from paasta_tools.utils import list_services
from paasta_tools.utils import load_system_paasta_config
from paasta_tools.utils import SPACER

try:
    import yelp_meteorite
except ImportError:
    yelp_meteorite = None

log = logging.getLogger(__name__)

CheckServiceReplication = Callable[
    [
        Arg(
            InstanceConfig_T,
            "instance_config",  # noqa: F821  # flake8 false-positive, these are not var references
        ),
        Arg(
            Dict[str, Dict[str, List[V1Pod]]],
            "pods_by_service_instance",  # noqa: F821  # flake8 false-positive
        ),
        Arg(Any, "replication_checker"),  # noqa: F821  # flake8 false-positive
        NamedArg(bool, "dry_run"),  # noqa: F821  # flake8 false-positive
    ],
    Optional[bool],
]


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=DEFAULT_SOA_DIR,
        help="define a different soa config directory",
    )
    parser.add_argument(
        "--crit",
        dest="under_replicated_crit_pct",
        type=float,
        default=10,
        help="The percentage of under replicated service instances past which "
        "the script will return a critical status",
    )
    parser.add_argument(
        "--min-count-critical",
        dest="min_count_critical",
        type=int,
        default=5,
        help="The script will not return a critical status if the number of "
        "under replicated service instances is below this number, even if the "
        "percentage is above the critical percentage.",
    )
    parser.add_argument(
        "service_instance_list",
        nargs="*",
        help="The list of service instances to check",
        metavar="SERVICE%sINSTANCE" % SPACER,
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", dest="verbose", default=False
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        dest="dry_run",
        help="Print Sensu alert events and metrics instead of sending them",
    )
    parser.add_argument(
        "--eks",
        help="This flag checks k8 services running on EKS",
        dest="eks",
        action="store_true",
        default=False,
    )
    options = parser.parse_args()

    return options


def check_services_replication(
    soa_dir: str,
    cluster: str,
    service_instances: Sequence[str],
    instance_type_class: Type[InstanceConfig_T],
    check_service_replication: CheckServiceReplication,
    replication_checker: ReplicationChecker,
    pods_by_service_instance: Dict[str, Dict[str, List[V1Pod]]],
    dry_run: bool = False,
) -> Tuple[int, int]:
    service_instances_set = set(service_instances)
    replication_statuses: List[bool] = []

    for service in list_services(soa_dir=soa_dir):
        service_config = PaastaServiceConfigLoader(service=service, soa_dir=soa_dir)
        for instance_config in service_config.instance_configs(
            cluster=cluster, instance_type_class=instance_type_class
        ):
            if (
                service_instances_set
                and f"{service}{SPACER}{instance_config.instance}"
                not in service_instances_set
            ):
                continue
            if instance_config.get_docker_image():
                is_well_replicated = check_service_replication(
                    instance_config=instance_config,
                    pods_by_service_instance=pods_by_service_instance,
                    replication_checker=replication_checker,
                    dry_run=dry_run,
                )
                if is_well_replicated is not None:
                    replication_statuses.append(is_well_replicated)

            else:
                log.debug(
                    "%s is not deployed. Skipping replication monitoring."
                    % instance_config.job_id
                )

    num_under_replicated = len(
        [status for status in replication_statuses if status is False]
    )
    return num_under_replicated, len(replication_statuses)


def emit_cluster_replication_metrics(
    pct_under_replicated: float,
    cluster: str,
    scheduler: str,
    dry_run: bool = False,
) -> None:
    metric_name = "paasta.pct_services_under_replicated"
    if dry_run:
        print(f"Would've sent value {pct_under_replicated} for metric '{metric_name}'")
    else:
        meteorite_dims = {"paasta_cluster": cluster, "scheduler": scheduler}
        gauge = yelp_meteorite.create_gauge(metric_name, meteorite_dims)
        gauge.set(pct_under_replicated)


def main(
    instance_type_class: Type[InstanceConfig_T],
    check_service_replication: CheckServiceReplication,
    namespace: str = None,
) -> None:
    args = parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    else:
        logging.basicConfig(level=logging.WARNING)

    system_paasta_config = load_system_paasta_config()
    cluster = system_paasta_config.get_cluster()
    replication_checker: ReplicationChecker

    timer = metrics_lib.system_timer(dimensions=dict(eks=args.eks, cluster=cluster))

    timer.start()

    if namespace:
        pods, nodes = get_kubernetes_pods_and_nodes(namespace=namespace)
        replication_checker = KubeSmartstackEnvoyReplicationChecker(
            nodes=nodes,
            system_paasta_config=system_paasta_config,
        )
    else:
        pods, nodes = get_kubernetes_pods_and_nodes()
        replication_checker = KubeSmartstackEnvoyReplicationChecker(
            nodes=nodes,
            system_paasta_config=system_paasta_config,
        )

    pods_by_service_instance = group_pods_by_service_instance(pods)

    count_under_replicated, total = check_services_replication(
        soa_dir=args.soa_dir,
        cluster=cluster,
        service_instances=args.service_instance_list,
        instance_type_class=instance_type_class,
        check_service_replication=check_service_replication,
        replication_checker=replication_checker,
        pods_by_service_instance=pods_by_service_instance,
        dry_run=args.dry_run,
    )
    pct_under_replicated = 0 if total == 0 else 100 * count_under_replicated / total
    if yelp_meteorite is not None:
        emit_cluster_replication_metrics(
            pct_under_replicated,
            cluster,
            scheduler="kubernetes",
            dry_run=args.dry_run,
        )

    exit_code = 0
    if (
        pct_under_replicated >= args.under_replicated_crit_pct
        and count_under_replicated >= args.min_count_critical
    ):
        log.critical(
            f"{pct_under_replicated}% of instances ({count_under_replicated}/{total}) "
            f"are under replicated (past {args.under_replicated_crit_pct} is critical)!"
        )
        exit_code = 2

    timer.stop(tmp_dimensions={"result": exit_code})
    logging.info(
        f"Stopping timer for {cluster} (eks={args.eks}) with result {exit_code}: {timer()}ms elapsed"
    )
    sys.exit(exit_code)


# XXX: is there a base class for the k8s clientlib models that we could use to type `obj`?
def set_local_vars_configuration_to_none(obj: Any, visited: Set[int] = None) -> None:
    """
    Recursive function to ensure that k8s clientlib objects are pickleable.

    Without this, k8s clientlib objects can't be used by multiprocessing functions
    as those pickle data to shuttle between processes.
    """
    if visited is None:
        visited = set()

    # Avoid infinite recursion for objects that have already been visited
    obj_id = id(obj)
    if obj_id in visited:
        return
    visited.add(obj_id)

    # if the object has the attribute, set it to None to essentially delete it
    if hasattr(obj, "local_vars_configuration"):
        setattr(obj, "local_vars_configuration", None)

    # recursively check attributes of the object
    if hasattr(obj, "__dict__"):
        for attr_name, attr_value in obj.__dict__.items():
            set_local_vars_configuration_to_none(attr_value, visited)

    # if the object is iterable/a collection, iterate over its elements
    elif isinstance(obj, (list, tuple, set)):
        for item in obj:
            set_local_vars_configuration_to_none(item, visited)
    elif isinstance(obj, dict):
        for value in obj.values():
            set_local_vars_configuration_to_none(value, visited)


def __fetch_pods(namespace: str) -> List[V1Pod]:
    kube_client = KubeClient()
    pods = get_all_pods(kube_client, namespace)
    for pod in pods:
        # this is pretty silly, but V1Pod cannot be pickled otherwise since the local_vars_configuration member
        # is not picklable - and pretty much every k8s model has this member ;_;
        set_local_vars_configuration_to_none(pod)
    return pods


def __get_all_pods_parallel(from_namespaces: Set[str]) -> List[V1Pod]:
    all_pods: List[V1Pod] = []
    with Pool() as pool:
        for pod_list in pool.imap_unordered(
            __fetch_pods,
            from_namespaces,
            chunksize=len(from_namespaces) // cast(int, cpu_count()),
        ):
            all_pods.extend(pod_list)
    return all_pods


def get_kubernetes_pods_and_nodes(
    namespace: Optional[str] = None,
) -> Tuple[List[V1Pod], List[V1Node]]:
    kube_client = KubeClient()

    if namespace:
        all_pods = get_all_pods(kube_client=kube_client, namespace=namespace)
    else:
        all_managed_namespaces = set(get_all_managed_namespaces(kube_client))
        all_pods = __get_all_pods_parallel(all_managed_namespaces)

    all_nodes = get_all_nodes(kube_client)

    return all_pods, all_nodes



#!/usr/bin/env python
# Copyright 2015-2018 Yelp Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./setup_tron_namespace.py service [service...] | --all

Deploy a namespace to the local Tron master from a service configuration file.
Reads from the soa_dir /nail/etc/services by default.

The script will load the service configuration file, generate a Tron configuration
file for it, and send the updated file to Tron.
"""
import argparse
import logging
import sys
from typing import Dict
from typing import List

import ruamel.yaml as yaml

from paasta_tools import spark_tools
from paasta_tools import tron_tools
from paasta_tools.kubernetes_tools import ensure_service_account
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.tron_tools import KUBERNETES_NAMESPACE
from paasta_tools.tron_tools import load_tron_service_config
from paasta_tools.tron_tools import MASTER_NAMESPACE
from paasta_tools.tron_tools import TronJobConfig
from paasta_tools.utils import load_system_paasta_config

log = logging.getLogger(__name__)


def parse_args():
    parser = argparse.ArgumentParser(
        description="Update the Tron namespace configuration for a service."
    )
    parser.add_argument("services", nargs="*", help="Services to update.")
    parser.add_argument(
        "-a",
        "--all",
        dest="all_namespaces",
        action="store_true",
        help="Update all available Tron namespaces.",
    )
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        metavar="SOA_DIR",
        default=tron_tools.DEFAULT_SOA_DIR,
        help="Use a different soa config directory",
    )
    parser.add_argument("-v", "--verbose", action="store_true", default=False)
    parser.add_argument("--dry-run", action="store_true", default=False)
    parser.add_argument(
        "--bulk-config-fetch",
        dest="bulk_config_fetch",
        action="store_true",
        default=False,
        help="Attempt to fetch all configs in bulk rather than one by one",
    )
    parser.add_argument(
        "--cluster",
        help="Cluster to read configs for. Defaults to the configuration in /etc/paasta",
        default=None,
    )
    args = parser.parse_args()
    return args


def ensure_service_accounts(job_configs: List[TronJobConfig]) -> None:
    # NOTE: these are lru_cache'd so it should be fine to call these for every service
    system_paasta_config = load_system_paasta_config()
    kube_client = KubeClient()

    for job in job_configs:
        for action in job.get_actions():
            if action.get_iam_role():
                ensure_service_account(
                    action.get_iam_role(),
                    namespace=KUBERNETES_NAMESPACE,
                    kube_client=kube_client,
                )
                # spark executors are special in that we want the SA to exist in two namespaces:
                # the tron namespace - for the spark driver (which will be created by the ensure_service_account() above)
                # and the spark namespace - for the spark executor (which we'll create below)
                if (
                    action.get_executor() == "spark"
                    # this should always be truthy, but let's be safe since this comes from SystemPaastaConfig
                    and action.get_spark_executor_iam_role()
                ):
                    # this kubeclient creation is lru_cache'd so it should be fine to call this for every spark action
                    spark_kube_client = KubeClient(
                        config_file=system_paasta_config.get_spark_kubeconfig()
                    )
                    # this will look quite similar to the above, but we're ensuring that a potentially different SA exists:
                    # this one is for the actual spark executors to use. if an iam_role is set, we'll use that, otherwise
                    # there's an executor-specifc default role just like there is for the drivers :)
                    ensure_service_account(
                        action.get_spark_executor_iam_role(),
                        namespace=spark_tools.SPARK_EXECUTOR_NAMESPACE,
                        kube_client=spark_kube_client,
                    )


def main() -> None:
    args = parse_args()
    log_level = logging.DEBUG if args.verbose else logging.INFO
    logging.basicConfig(level=log_level)

    if not args.cluster:
        args.cluster = tron_tools.load_tron_config().get_cluster_name()

    if args.all_namespaces:
        if args.services:
            log.error("Do not pass service names with --all flag")
            sys.exit(1)

        try:
            services = tron_tools.get_tron_namespaces(
                cluster=args.cluster, soa_dir=args.soa_dir
            )
        except Exception:
            log.exception("Failed to list tron namespaces:")
            sys.exit(1)
    else:
        services = args.services

    if not services:
        log.warning("No namespaces found")
        sys.exit(0)

    if not args.dry_run:
        client = tron_tools.get_tron_client()

    updated = []
    failed = []
    skipped = []

    master_config = tron_tools.create_complete_master_config(
        cluster=args.cluster, soa_dir=args.soa_dir
    )
    if args.dry_run:
        log.info(f"Would update {MASTER_NAMESPACE} to:")
        log.info(f"{master_config}")
        updated.append(MASTER_NAMESPACE)
    else:
        try:
            if client.update_namespace(MASTER_NAMESPACE, master_config):
                updated.append(MASTER_NAMESPACE)
                log.debug(f"Updated {MASTER_NAMESPACE}")
            else:
                skipped.append(MASTER_NAMESPACE)
                log.debug(f"Skipped {MASTER_NAMESPACE}")
        except Exception:
            failed.append(MASTER_NAMESPACE)
            log.exception(f"Error while updating {MASTER_NAMESPACE}:")

    k8s_enabled_for_cluster = (
        yaml.safe_load(master_config).get("k8s_options", {}).get("enabled", False)
    )
    new_configs: Dict[str, str] = {}  # service -> new_config
    for service in sorted(services):
        try:
            new_config = tron_tools.create_complete_config(
                cluster=args.cluster,
                service=service,
                soa_dir=args.soa_dir,
                k8s_enabled=k8s_enabled_for_cluster,
                dry_run=args.dry_run,
            )
            new_configs[service] = new_config
            if args.dry_run:
                log.info(f"Would update {service} to:")
                log.info(f"{new_config}")
                updated.append(service)
            else:
                # PaaSTA will not necessarily have created the SAs we want to use
                # ...so let's go ahead and create them!
                job_configs = load_tron_service_config(
                    service=service,
                    cluster=args.cluster,
                    load_deployments=False,
                    soa_dir=args.soa_dir,
                    # XXX: we can remove for_validation now that we've refactored how service account stuff works
                    for_validation=False,
                )
                ensure_service_accounts(job_configs)
                if not args.bulk_config_fetch:
                    if client.update_namespace(service, new_config):
                        updated.append(service)
                        log.debug(f"Updated {service}")
                    else:
                        skipped.append(service)
                        log.debug(f"Skipped {service}")

        except Exception:
            if args.bulk_config_fetch:
                # service account creation should be the only action that can throw if this flag is true,
                # so we can safely assume that's what happened here in the log message
                log.exception(
                    f"Failed to create service account for {service} (will skip reconfiguring):"
                )

                # since service account creation failed, we want to skip reconfiguring this service
                # as the new config will likely fail due to the missing service account - even though
                # the rest of the config is valid
                new_configs.pop(service, None)
            else:
                log.exception(f"Update for {service} failed:")

            # NOTE: this happens for both ways of updating (bulk fetch and JIT fetch)
            # since we need to print out what failed in either case
            failed.append(service)

    if args.dry_run and args.bulk_config_fetch:
        updated_namespaces = client.update_namespaces(new_configs)

        if updated_namespaces:
            updated = list(updated_namespaces.keys())
            log.debug(f"Updated {updated}")

        if updated_namespaces != new_configs.keys():
            skipped = list(set(new_configs.keys()) - set(updated_namespaces.keys()))
            log.debug(f"Skipped {skipped}")

    skipped_report = skipped if args.verbose else len(skipped)
    log.info(
        f"Updated following namespaces: {updated}, "
        f"failed: {failed}, skipped: {skipped_report}"
    )

    sys.exit(1 if failed else 0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from unittest import mock

import pytest
from kubernetes.client import V1Deployment
from kubernetes.client import V1StatefulSet
from pytest import fixture
from pytest import raises

from paasta_tools.cleanup_kubernetes_jobs import cleanup_unused_apps
from paasta_tools.cleanup_kubernetes_jobs import DontKillEverythingError
from paasta_tools.cleanup_kubernetes_jobs import main
from paasta_tools.eks_tools import EksDeploymentConfig
from paasta_tools.kubernetes.application.controller_wrappers import DeploymentWrapper
from paasta_tools.kubernetes.application.controller_wrappers import StatefulSetWrapper
from paasta_tools.kubernetes_tools import KubernetesDeploymentConfig


@fixture
def fake_deployment():
    fake_deployment = V1Deployment(
        metadata=mock.Mock(
            namespace="paastasvc-service",
            labels={
                "yelp.com/paasta_service": "service",
                "yelp.com/paasta_instance": "instance-1",
                "yelp.com/paasta_git_sha": "1234",
                "yelp.com/paasta_config_sha": "1234",
                "paasta.yelp.com/service": "service",
                "paasta.yelp.com/instance": "instance-1",
                "paasta.yelp.com/git_sha": "1234",
                "paasta.yelp.com/config_sha": "1234",
                "paasta.yelp.com/managed": "true",
            },
        ),
        spec=mock.Mock(replicas=0),
        status=mock.Mock(ready_replicas=0),
    )
    type(fake_deployment.metadata).name = mock.PropertyMock(
        return_value="service-instance-1"
    )
    return fake_deployment


@fixture
def fake_stateful_set():
    fake_stateful_set = V1StatefulSet(
        metadata=mock.Mock(
            namespace="paasta",
            labels={
                "yelp.com/paasta_service": "service",
                "yelp.com/paasta_instance": "instance-2",
                "yelp.com/paasta_git_sha": "1234",
                "yelp.com/paasta_config_sha": "1234",
                "paasta.yelp.com/service": "service",
                "paasta.yelp.com/instance": "instance-2",
                "paasta.yelp.com/git_sha": "1234",
                "paasta.yelp.com/config_sha": "1234",
                "paasta.yelp.com/managed": "true",
            },
        ),
        spec=mock.Mock(replicas=0),
    )
    type(fake_stateful_set.metadata).name = (
        mock.PropertyMock(return_value="service-instance-2"),
    )
    return fake_stateful_set


@fixture
def invalid_app():
    invalid_app = V1Deployment(
        metadata=mock.Mock(namespace="paasta", labels={}), spec=mock.Mock(replicas=0)
    )
    type(invalid_app.metadata).name = (mock.PropertyMock(return_value="invalid_app"),)
    return invalid_app


def fake_instance_config(
    cluster, service, instance, soa_dir="soa_dir", load_deployments=False
):
    fake_instance_config = KubernetesDeploymentConfig(
        service,
        instance,
        cluster,
        {
            "port": None,
            "monitoring": {},
            "deploy": {"pipeline": [{"step": "default"}]},
            "data": {},
            "smartstack": {},
            "dependencies": {},
            "cpus": 0.1,
            "mem": 100,
            "min_instances": 1,
            "max_instances": 10,
            "deploy_group": "prod.main",
            "autoscaling": {"setpoint": 0.7},
        },
        {
            "docker_image": "services-compute-infra-test-service:paasta-5b861b3bd42ef9674d3ca04a1259c79eddb71694",
            "git_sha": "5b861b3bd42ef9674d3ca04a1259c79eddb71694",
            "image_version": None,
            "desired_state": "start",
            "force_bounce": None,
        },
        soa_dir,
    )
    return fake_instance_config


def fake_eks_instance_config(
    cluster, service, instance, soa_dir="soa_dir", load_deployments=False
):
    fake_eks_instance_config = EksDeploymentConfig(
        service,
        instance,
        cluster,
        {
            "port": None,
            "monitoring": {},
            "deploy": {"pipeline": [{"step": "default"}]},
            "data": {},
            "smartstack": {},
            "dependencies": {},
            "cpus": 0.1,
            "mem": 100,
            "min_instances": 1,
            "max_instances": 10,
            "deploy_group": "prod.main",
            "autoscaling": {"setpoint": 0.7},
        },
        {
            "docker_image": "services-compute-infra-test-service:paasta-5b861b3bd42ef9674d3ca04a1259c79eddb71694",
            "git_sha": "5b861b3bd42ef9674d3ca04a1259c79eddb71694",
            "image_version": None,
            "desired_state": "start",
            "force_bounce": None,
        },
        soa_dir,
    )
    return fake_eks_instance_config


def get_fake_instances(self, with_limit: bool = True) -> int:
    return self.config_dict.get("max_instances", None)


def test_main(fake_deployment, fake_stateful_set, invalid_app):
    soa_dir = "paasta_maaaachine"
    cluster = "maaaachine_cluster"
    with mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.cleanup_unused_apps", autospec=True
    ) as cleanup_patch, mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.load_system_paasta_config", autospec=True
    ) as load_config_patch:
        load_config_patch.return_value.get_cluster.return_value = "fake_cluster"
        main(("--soa-dir", soa_dir, "--cluster", cluster))
        cleanup_patch.assert_called_once_with(
            soa_dir, cluster, kill_threshold=0.5, force=False, eks=False
        )


def test_list_apps(fake_deployment, fake_stateful_set, invalid_app):
    mock_kube_client = mock.MagicMock()
    with mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubeClient",
        return_value=mock_kube_client,
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.get_services_for_cluster",
        return_value={},
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.alert_state_change", autospec=True
    ) as mock_alert_state_change:
        mock_alert_state_change.__enter__ = mock.Mock(return_value=(mock.Mock(), None))
        mock_alert_state_change.__exit__ = mock.Mock(return_value=None)
        cleanup_unused_apps("soa_dir", "fake cluster", kill_threshold=1, force=False)
        assert (
            mock_kube_client.deployments.list_deployment_for_all_namespaces.call_count
            == 1
        )
        assert (
            mock_kube_client.deployments.list_stateful_set_for_all_namespaces.call_count
            == 1
        )


@pytest.mark.parametrize(
    "eks_flag",
    [
        (False),
        (True),
    ],
)
def test_cleanup_unused_apps(eks_flag, fake_deployment, fake_stateful_set, invalid_app):
    mock_kube_client = mock.MagicMock()
    with mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubeClient",
        return_value=mock_kube_client,
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.list_all_applications",
        return_value={("service", "instance-1"): [DeploymentWrapper(fake_deployment)]},
        autospec=True,
    ), mock.patch(
        "paasta_tools.kubernetes_tools.load_kubernetes_service_config_no_cache",
        autospec=True,
        side_effect=fake_instance_config,
    ), mock.patch(
        "paasta_tools.eks_tools.load_eks_service_config_no_cache",
        autospec=True,
        side_effect=fake_eks_instance_config,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubernetesDeploymentConfig.get_instances",
        side_effect=get_fake_instances,
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.get_services_for_cluster",
        return_value={},
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.alert_state_change", autospec=True
    ) as mock_alert_state_change:
        mock_alert_state_change.__enter__ = mock.Mock(return_value=(mock.Mock(), None))
        mock_alert_state_change.__exit__ = mock.Mock(return_value=None)
        cleanup_unused_apps(
            "soa_dir", "fake cluster", kill_threshold=1, force=False, eks=eks_flag
        )
        assert mock_kube_client.deployments.delete_namespaced_deployment.call_count == 1


@pytest.mark.parametrize(
    "eks_flag",
    [
        (False),
        (True),
    ],
)
def test_cleanup_unused_apps_in_multiple_namespaces(
    eks_flag, fake_deployment, fake_stateful_set, invalid_app
):
    mock_kube_client = mock.MagicMock()
    fake_deployment2 = deepcopy(fake_deployment)
    fake_deployment2.metadata.namespace = "paastasvc-blah"
    fake_deployment.status.ready_replicas = 10
    fake_deployment2.status.ready_replicas = 3
    with mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubeClient",
        return_value=mock_kube_client,
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.list_all_applications",
        return_value={
            ("service", "instance-1"): [
                DeploymentWrapper(fake_deployment),
                DeploymentWrapper(fake_deployment2),
            ]
        },
        autospec=True,
    ), mock.patch(
        "paasta_tools.kubernetes_tools.load_kubernetes_service_config_no_cache",
        autospec=True,
        side_effect=fake_instance_config,
    ), mock.patch(
        "paasta_tools.eks_tools.load_eks_service_config_no_cache",
        autospec=True,
        side_effect=fake_eks_instance_config,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.get_services_for_cluster",
        return_value={("service", "instance-1")},
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubernetesDeploymentConfig.get_instances",
        side_effect=get_fake_instances,
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.alert_state_change", autospec=True
    ) as mock_alert_state_change:
        mock_alert_state_change.__enter__ = mock.Mock(return_value=(mock.Mock(), None))
        mock_alert_state_change.__exit__ = mock.Mock(return_value=None)
        cleanup_unused_apps(
            "soa_dir", "fake cluster", kill_threshold=2, force=False, eks=eks_flag
        )
        assert mock_kube_client.deployments.delete_namespaced_deployment.call_count == 1


@pytest.mark.parametrize(
    "eks_flag",
    [
        (False),
        (True),
    ],
)
def test_cleanup_unused_apps_does_not_delete(
    eks_flag, fake_deployment, fake_stateful_set, invalid_app
):
    mock_kube_client = mock.MagicMock()
    with mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubeClient",
        return_value=mock_kube_client,
        autospec=True,
    ), mock.patch(
        "paasta_tools.kubernetes_tools.load_kubernetes_service_config_no_cache",
        autospec=True,
        side_effect=fake_instance_config,
    ), mock.patch(
        "paasta_tools.eks_tools.load_eks_service_config_no_cache",
        autospec=True,
        side_effect=fake_eks_instance_config,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.list_all_applications",
        return_value={("service", "instance-1"): [DeploymentWrapper(fake_deployment)]},
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubernetesDeploymentConfig.get_instances",
        side_effect=get_fake_instances,
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.get_services_for_cluster",
        return_value={("service", "instance-1"), ("service", "instance-2")},
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.alert_state_change", autospec=True
    ) as mock_alert_state_change:
        mock_alert_state_change.__enter__ = mock.Mock(return_value=(mock.Mock(), None))
        mock_alert_state_change.__exit__ = mock.Mock(return_value=None)
        cleanup_unused_apps(
            "soa_dir", "fake cluster", kill_threshold=1, force=False, eks=eks_flag
        )
        assert mock_kube_client.deployments.delete_namespaced_deployment.call_count == 0


@pytest.mark.parametrize(
    "eks_flag",
    [
        (False),
        (True),
    ],
)
def test_cleanup_unused_apps_does_not_delete_bouncing_apps(
    eks_flag, fake_deployment, fake_stateful_set, invalid_app
):
    mock_kube_client = mock.MagicMock()
    fake_deployment2 = deepcopy(fake_deployment)
    fake_deployment2.status.ready_replicas = 3
    fake_deployment.status.ready_replicas = 10
    fake_deployment.metadata.namespace = "paastasvc-blah"
    with mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubeClient",
        return_value=mock_kube_client,
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.list_all_applications",
        return_value={
            ("service", "instance-1"): [
                DeploymentWrapper(fake_deployment),
                DeploymentWrapper(fake_deployment2),
            ]
        },
        autospec=True,
    ), mock.patch(
        "paasta_tools.kubernetes_tools.load_kubernetes_service_config_no_cache",
        autospec=True,
        side_effect=fake_instance_config,
    ), mock.patch(
        "paasta_tools.eks_tools.load_eks_service_config_no_cache",
        autospec=True,
        side_effect=fake_eks_instance_config,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.get_services_for_cluster",
        return_value={("service", "instance-1")},
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubernetesDeploymentConfig.get_instances",
        side_effect=get_fake_instances,
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.alert_state_change", autospec=True
    ) as mock_alert_state_change:
        mock_alert_state_change.__enter__ = mock.Mock(return_value=(mock.Mock(), None))
        mock_alert_state_change.__exit__ = mock.Mock(return_value=None)
        cleanup_unused_apps(
            "soa_dir", "fake cluster", kill_threshold=2, force=False, eks=eks_flag
        )
        assert mock_kube_client.deployments.delete_namespaced_deployment.call_count == 0


@pytest.mark.parametrize(
    "eks_flag",
    [
        (False),
        (True),
    ],
)
def test_cleanup_unused_apps_does_not_delete_recently_created_apps(
    eks_flag, fake_deployment, fake_stateful_set, invalid_app
):
    mock_kube_client = mock.MagicMock()
    fake_deployment.status.ready_replicas = 10
    fake_deployment.metadata.namespace = "paastasvc-blah"
    with mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubeClient",
        return_value=mock_kube_client,
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.list_all_applications",
        return_value={
            ("service", "instance-1"): [
                DeploymentWrapper(fake_deployment),
            ]
        },
        autospec=True,
    ), mock.patch(
        "paasta_tools.kubernetes_tools.load_kubernetes_service_config_no_cache",
        autospec=True,
        side_effect=fake_instance_config,
    ), mock.patch(
        "paasta_tools.eks_tools.load_eks_service_config_no_cache",
        autospec=True,
        side_effect=fake_eks_instance_config,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.get_services_for_cluster",
        return_value={("service", "instance-1")},
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubernetesDeploymentConfig.get_instances",
        side_effect=get_fake_instances,
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.alert_state_change", autospec=True
    ) as mock_alert_state_change:
        mock_alert_state_change.__enter__ = mock.Mock(return_value=(mock.Mock(), None))
        mock_alert_state_change.__exit__ = mock.Mock(return_value=None)
        cleanup_unused_apps(
            "soa_dir", "fake cluster", kill_threshold=2, force=False, eks=eks_flag
        )
        assert mock_kube_client.deployments.delete_namespaced_deployment.call_count == 0


@pytest.mark.parametrize(
    "eks_flag",
    [
        (False),
        (True),
    ],
)
def test_cleanup_unused_apps_dont_kill_everything(
    eks_flag, fake_deployment, fake_stateful_set, invalid_app
):
    mock_kube_client = mock.MagicMock()
    with mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubeClient",
        return_value=mock_kube_client,
        autospec=True,
    ), mock.patch(
        "paasta_tools.kubernetes_tools.load_kubernetes_service_config_no_cache",
        autospec=True,
        side_effect=fake_instance_config,
    ), mock.patch(
        "paasta_tools.eks_tools.load_eks_service_config_no_cache",
        autospec=True,
        side_effect=fake_eks_instance_config,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.list_all_applications",
        return_value={("service", "instance-1"): [DeploymentWrapper(fake_deployment)]},
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubernetesDeploymentConfig.get_instances",
        side_effect=get_fake_instances,
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.get_services_for_cluster",
        return_value={},
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.alert_state_change", autospec=True
    ) as mock_alert_state_change:
        mock_alert_state_change.__enter__ = mock.Mock(return_value=(mock.Mock(), None))
        mock_alert_state_change.__exit__ = mock.Mock(return_value=None)
        with raises(DontKillEverythingError):
            cleanup_unused_apps(
                "soa_dir", "fake_cluster", kill_threshold=0, force=False, eks=eks_flag
            )
        assert mock_kube_client.deployments.delete_namespaced_deployment.call_count == 0


@pytest.mark.parametrize(
    "eks_flag",
    [
        (False),
        (True),
    ],
)
def test_cleanup_unused_apps_dont_kill_statefulsets(
    eks_flag, fake_deployment, fake_stateful_set, invalid_app
):
    mock_kube_client = mock.MagicMock()
    with mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubeClient",
        return_value=mock_kube_client,
        autospec=True,
    ), mock.patch(
        "paasta_tools.kubernetes_tools.load_kubernetes_service_config_no_cache",
        autospec=True,
        side_effect=fake_instance_config,
    ), mock.patch(
        "paasta_tools.eks_tools.load_eks_service_config_no_cache",
        autospec=True,
        side_effect=fake_eks_instance_config,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.list_all_applications",
        return_value={
            ("service", "instance-2"): [
                StatefulSetWrapper(fake_stateful_set),
                StatefulSetWrapper(fake_stateful_set),
            ]
        },
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.get_services_for_cluster",
        return_value={("service", "instance-2")},
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.alert_state_change", autospec=True
    ) as mock_alert_state_change:
        mock_alert_state_change.__enter__ = mock.Mock(return_value=(mock.Mock(), None))
        mock_alert_state_change.__exit__ = mock.Mock(return_value=None)
        cleanup_unused_apps(
            "soa_dir", "fake_cluster", kill_threshold=0.5, force=False, eks=eks_flag
        )
        assert mock_kube_client.deployments.delete_namespaced_deployment.call_count == 0


@pytest.mark.parametrize(
    "eks_flag",
    [
        (False),
        (True),
    ],
)
def test_cleanup_unused_apps_force(
    eks_flag, fake_deployment, fake_stateful_set, invalid_app
):
    mock_kube_client = mock.MagicMock()
    with mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubeClient",
        return_value=mock_kube_client,
        autospec=True,
    ), mock.patch(
        "paasta_tools.kubernetes_tools.load_kubernetes_service_config_no_cache",
        autospec=True,
        side_effect=fake_instance_config,
    ), mock.patch(
        "paasta_tools.eks_tools.load_eks_service_config_no_cache",
        autospec=True,
        side_effect=fake_eks_instance_config,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.list_all_applications",
        return_value={("service", "instance-1"): [DeploymentWrapper(fake_deployment)]},
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubernetesDeploymentConfig.get_instances",
        side_effect=get_fake_instances,
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.get_services_for_cluster",
        return_value={},
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.alert_state_change", autospec=True
    ) as mock_alert_state_change:
        mock_alert_state_change.__enter__ = mock.Mock(return_value=(mock.Mock(), None))
        mock_alert_state_change.__exit__ = mock.Mock(return_value=None)
        cleanup_unused_apps(
            "soa_dir", "fake_cluster", kill_threshold=0, force=True, eks=eks_flag
        )
        assert mock_kube_client.deployments.delete_namespaced_deployment.call_count == 1


@pytest.mark.parametrize(
    "eks_flag",
    [
        (False),
        (True),
    ],
)
def test_cleanup_unused_apps_ignore_invalid_apps(
    eks_flag, fake_deployment, fake_stateful_set, invalid_app
):
    mock_kube_client = mock.MagicMock()
    with mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.KubeClient",
        return_value=mock_kube_client,
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.get_services_for_cluster",
        return_value={},
        autospec=True,
    ), mock.patch(
        "paasta_tools.cleanup_kubernetes_jobs.alert_state_change", autospec=True
    ) as mock_alert_state_change:
        mock_alert_state_change.__enter__ = mock.Mock(return_value=(mock.Mock(), None))
        mock_alert_state_change.__exit__ = mock.Mock(return_value=None)
        mock_kube_client.deployments.list_namespaced_deployment.return_value = (
            mock.MagicMock(items=[invalid_app])
        )
        cleanup_unused_apps(
            "soa_dir", "fake_cluster", kill_threshold=0, force=True, eks=eks_flag
        )
        assert mock_kube_client.deployments.delete_namespaced_deployment.call_count == 0



#!/usr/bin/env python3.10
import ast
import sys


class MockChecker(ast.NodeVisitor):
    def __init__(self):
        self.errors = 0
        self.init_module_imports()

    def init_module_imports(self):
        self.imported_patch = False
        self.imported_mock = False

    def check_files(self, files):
        for file in files:
            self.check_file(file)

    def check_file(self, filename):
        self.current_filename = filename
        try:
            with open(filename, "r") as fd:
                try:
                    file_ast = ast.parse(fd.read())
                except SyntaxError as error:
                    print("SyntaxError on file %s:%d" % (filename, error.lineno))
                    return
        except IOError:
            print("Error opening filename: %s" % filename)
            return
        self.init_module_imports()
        self.visit(file_ast)

    def _call_uses_patch(self, node):
        try:
            return node.func.id == "patch"
        except AttributeError:
            return False

    def _call_uses_mock_patch(self, node):
        try:
            return node.func.value.id == "mock" and node.func.attr == "patch"
        except AttributeError:
            return False

    def visit_Import(self, node):
        if [name for name in node.names if "mock" == name.name]:
            self.imported_mock = True

    def visit_ImportFrom(self, node):
        if node.module == "mock" and (
            name for name in node.names if "patch" == name.name
        ):
            self.imported_patch = True

    def visit_Call(self, node):
        try:
            if (self.imported_patch and self._call_uses_patch(node)) or (
                self.imported_mock and self._call_uses_mock_patch(node)
            ):
                if not any(
                    [keyword for keyword in node.keywords if keyword.arg == "autospec"]
                ):
                    print(
                        "%s:%d: Found a mock without an autospec!"
                        % (self.current_filename, node.lineno)
                    )
                    self.errors += 1
        except AttributeError:
            pass
        self.generic_visit(node)


def main(filenames):
    checker = MockChecker()
    checker.check_files(filenames)
    if checker.errors == 0:
        sys.exit(0)
    else:
        print("You probably meant to specify 'autospec=True' in these tests.")
        print("If you really don't want to, specify 'autospec=None'")
        sys.exit(1)


if __name__ == "__main__":
    main(sys.argv[1:])



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import glob

from pkg_resources import yield_lines
from setuptools import find_packages
from setuptools import setup

from paasta_tools import __version__


def get_install_requires():
    with open("requirements-minimal.txt", "r") as f:
        minimal_reqs = list(yield_lines(f.read()))

    return minimal_reqs


setup(
    name="paasta-tools",
    version=__version__,
    provides=["paasta_tools"],
    author="Compute Infrastructure @ Yelp",
    author_email="compute-infra@yelp.com",
    description="Tools for Yelps SOA infrastructure",
    packages=find_packages(exclude=("tests*", "scripts*")),
    include_package_data=True,
    python_requires=">=3.9.0",
    install_requires=get_install_requires(),
    scripts=[
        "paasta_tools/apply_external_resources.py",
        "paasta_tools/check_autoscaler_max_instances.py",
        "paasta_tools/check_cassandracluster_services_replication.py",
        "paasta_tools/check_flink_services_health.py",
        "paasta_tools/check_kubernetes_api.py",
        "paasta_tools/check_kubernetes_services_replication.py",
        "paasta_tools/check_oom_events.py",
        "paasta_tools/check_spark_jobs.py",
        "paasta_tools/cleanup_kubernetes_cr.py",
        "paasta_tools/cleanup_kubernetes_crd.py",
        "paasta_tools/cleanup_kubernetes_jobs.py",
        "paasta_tools/cli/paasta_tabcomplete.sh",
        "paasta_tools/delete_kubernetes_deployments.py",
        "paasta_tools/generate_all_deployments",
        "paasta_tools/generate_deployments_for_service.py",
        "paasta_tools/generate_services_file.py",
        "paasta_tools/generate_services_yaml.py",
        "paasta_tools/generate_authenticating_services.py",
        "paasta_tools/kubernetes/bin/kubernetes_remove_evicted_pods.py",
        "paasta_tools/kubernetes/bin/paasta_cleanup_remote_run_resources.py",
        "paasta_tools/kubernetes/bin/paasta_cleanup_stale_nodes.py",
        "paasta_tools/kubernetes/bin/paasta_secrets_sync.py",
        "paasta_tools/paasta_deploy_tron_jobs",
        "paasta_tools/paasta_execute_docker_command.py",
        "paasta_tools/setup_istio_mesh.py",
        "paasta_tools/setup_kubernetes_cr.py",
        "paasta_tools/setup_kubernetes_crd.py",
        "paasta_tools/setup_kubernetes_internal_crd.py",
        "paasta_tools/setup_kubernetes_job.py",
        "paasta_tools/setup_prometheus_adapter_config.py",
        "paasta_tools/synapse_srv_namespaces_fact.py",
    ]
    + glob.glob("paasta_tools/contrib/*.sh")
    + glob.glob("paasta_tools/contrib/[!_]*.py"),
    entry_points={
        "console_scripts": [
            "paasta=paasta_tools.cli.cli:main",
            "paasta-api=paasta_tools.api.api:main",
            "paasta-deployd=paasta_tools.deployd.master:main",
            "paasta-fsm=paasta_tools.cli.fsm_cmd:main",
            "paasta_prune_completed_pods=paasta_tools.prune_completed_pods:main",
            "paasta_cleanup_tron_namespaces=paasta_tools.cleanup_tron_namespaces:main",
            "paasta_cleanup_expired_autoscaling_overrides=paasta_tools.cleanup_expired_autoscaling_overrides:main",
            "paasta_list_kubernetes_service_instances=paasta_tools.list_kubernetes_service_instances:main",
            "paasta_list_tron_namespaces=paasta_tools.list_tron_namespaces:main",
            "paasta_setup_tron_namespace=paasta_tools.setup_tron_namespace:main",
            "paasta_oom_logger=paasta_tools.oom_logger:main",
            "paasta_broadcast_log=paasta_tools.broadcast_log_to_services:main",
            "paasta_dump_locally_running_services=paasta_tools.dump_locally_running_services:main",
            "paasta_habitat_fixer=paasta_tools.contrib.habitat_fixer:main",
        ],
        "paste.app_factory": ["paasta-api-config=paasta_tools.api.api:make_app"],
    },
)



#!/usr/bin/env python
# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Contains methods used by the paasta client to build a docker image."""
import argparse
import os
import sys
from typing import Optional

from paasta_tools.cli.cmds.check import makefile_responds_to
from paasta_tools.cli.utils import validate_service_name
from paasta_tools.utils import _log
from paasta_tools.utils import _log_audit
from paasta_tools.utils import _run
from paasta_tools.utils import build_docker_tag
from paasta_tools.utils import DEFAULT_SOA_DIR
from paasta_tools.utils import get_username


def add_subparser(subparsers: argparse._SubParsersAction) -> None:
    list_parser = subparsers.add_parser(
        "cook-image",
        description="Calls 'make cook-image' as part of the PaaSTA contract",
        help=(
            "'paasta cook-image' calls 'make cook-image' as part of the PaaSTA contract.\n\n"
            "The PaaSTA contract specifies that a service MUST respond to 'cook-image' and produce "
            "a docker image as a result. This command is often run as part of the normal build pipeline "
            "('paasta itest'), or via a 'paasta local-run --build'."
        ),
        epilog="This command assumes that the Makefile is in the current working directory.",
    )
    list_parser.add_argument(
        "-s",
        "--service",
        help=(
            "Build docker image for this service. Leading "
            '"services-", as included in a Jenkins job name, '
            "will be stripped."
        ),
        # strip any potential trailing / for folks tab-completing directories
        type=lambda x: x.rstrip("/"),
        required=True,
    )
    list_parser.add_argument(
        "-y",
        "--yelpsoa-config-root",
        dest="yelpsoa_config_root",
        help="A directory from which yelpsoa-configs should be read from",
        default=DEFAULT_SOA_DIR,
    )
    list_parser.add_argument(
        "-c",
        "--commit",
        help="Git sha used to construct tag for built image",
    )
    list_parser.add_argument(
        "--image-version",
        type=str,
        required=False,
        default=None,
        help="Extra version metadata used to construct tag for built image",
    )
    list_parser.set_defaults(command=paasta_cook_image)


def paasta_cook_image(
    args: Optional[argparse.Namespace],
    service: Optional[str] = None,
    soa_dir: Optional[str] = None,
) -> int:
    """Build a docker image"""
    if not service:
        if args is None:
            print(
                "ERROR: No arguments or service passed to cook-image - unable to determine what service to cook an image for",
                file=sys.stderr,
            )
            return 1
        service = args.service
    if service and service.startswith("services-"):
        service = service.split("services-", 1)[1]
    if not soa_dir:
        if args is None:
            print(
                "ERROR: No arguments or soadir passed to cook-image - unable to determine where to look for soa-configs",
                file=sys.stderr,
            )
            return 1
        soa_dir = args.yelpsoa_config_root

    validate_service_name(service, soa_dir)

    run_env = os.environ.copy()
    if args is not None and args.commit is not None:
        # if we're given a commit, we're likely being called by Jenkins or someone
        # trying to push the cooked image to our registry - as such, we should tag
        # the cooked image as `paasta itest` would.
        tag = build_docker_tag(service, args.commit, args.image_version)
    else:
        default_tag = "paasta-cook-image-{}-{}".format(service, get_username())
        tag = run_env.get("DOCKER_TAG", default_tag)
    run_env["DOCKER_TAG"] = tag

    if not makefile_responds_to("cook-image"):
        print(
            "ERROR: local-run now requires a cook-image target to be present in the Makefile. See "
            "http://paasta.readthedocs.io/en/latest/about/contract.html.",
            file=sys.stderr,
        )
        return 1

    try:
        cmd = "make cook-image"
        returncode, output = _run(
            cmd,
            env=run_env,
            log=True,
            component="build",
            service=service,
            loglevel="debug",
        )
        if returncode != 0:
            _log(
                service=service,
                line="ERROR: make cook-image failed for %s." % service,
                component="build",
                level="event",
            )
        else:
            action_details = {"tag": tag}
            _log_audit(
                action="cook-image", action_details=action_details, service=service
            )
        return returncode

    except KeyboardInterrupt:
        print("\nProcess interrupted by the user. Cancelling.", file=sys.stderr)
        return 2



#!/usr/bin/env python
# Copyright 2015-2018 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Usage: ./setup_kubernetes_services.py <service.instance>

Command line options:

- -c, --cluster: PAASTA cluster
- -v, --verbose: Verbose output
"""
import argparse
import logging
import os
import sys
import time
from functools import partial
from typing import AbstractSet
from typing import Iterator
from typing import Mapping
from typing import Set

import kubernetes.client as k8s

from paasta_tools import yaml_tools as yaml
from paasta_tools.kubernetes_tools import ensure_namespace
from paasta_tools.kubernetes_tools import KubeClient
from paasta_tools.kubernetes_tools import limit_size_with_hash
from paasta_tools.kubernetes_tools import paasta_prefixed
from paasta_tools.kubernetes_tools import registration_label
from paasta_tools.kubernetes_tools import sanitise_kubernetes_name
from paasta_tools.utils import DEFAULT_SOA_DIR


log = logging.getLogger(__name__)

UNIFIED_K8S_SVC_NAME = "paasta-routing"
UNIFIED_SVC_PORT = 1337
PAASTA_SVC_PORT = 8888
PAASTA_NAMESPACE = "paasta"
ANNOTATIONS = {paasta_prefixed("managed_by"): "setup_istio_mesh"}


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Creates Kubernetes services.")
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        dest="verbose",
        default=False,
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        dest="dry_run",
        default=False,
    )
    parser.add_argument(
        "-l",
        "--rate-limit",
        dest="rate_limit",
        default=0,
        metavar="LIMIT",
        type=float,
        help="Maximum number of write calls to k8s per second. Default is 0 (no limit).",
    )
    parser.add_argument(
        "-d",
        "--soa-dir",
        dest="soa_dir",
        default=DEFAULT_SOA_DIR,
        metavar="LIMIT",
        type=str,
        help=f"Directory with service declarations. Default is {DEFAULT_SOA_DIR}",
    )
    args = parser.parse_args()
    return args


def load_smartstack_namespaces(soa_dir: str = DEFAULT_SOA_DIR) -> Mapping:
    namespaces = {}

    _, dirs, _ = next(os.walk(soa_dir))
    for dir in dirs:
        file_path = f"{soa_dir}/{dir}/smartstack.yaml"
        if not os.path.isfile(file_path):
            continue
        try:
            with open(file_path) as f:
                svc_namespaces = yaml.safe_load(f)
                for (ns, details) in svc_namespaces.items():
                    namespaces[f"{dir}.{ns}"] = details
        except Exception as err:
            log.warning(f"Failed to load namespaces for {dir}: {err}")

    return namespaces


def sanitise_kubernetes_service_name(name: str) -> str:
    return limit_size_with_hash(sanitise_kubernetes_name(name).replace(".", "---"))


def get_existing_kubernetes_service_names(kube_client: KubeClient) -> Set[str]:
    service_objects = kube_client.core.list_namespaced_service(PAASTA_NAMESPACE)

    return {
        item.metadata.name
        for item in service_objects.items
        if item.metadata.annotations
        if item.metadata.annotations.get(paasta_prefixed("managed_by"))
        == "setup_istio_mesh"
    }


def get_existing_kubernetes_virtual_services(kube_client: KubeClient) -> Set[str]:
    virtual_service_objects = kube_client.custom.list_namespaced_custom_object(
        "networking.istio.io", "v1beta1", PAASTA_NAMESPACE, "virtualservices"
    )

    return {item["metadata"]["name"] for item in virtual_service_objects["items"]}


def setup_paasta_routing(kube_client: KubeClient, namespaces: Mapping) -> Iterator:
    # Add smartstack ports for routing, Clients can connect to this
    # Directly without need of setting x-yelp-svc header
    # Add port 1337 for envoy unified listener.
    # Clients can connect to this listenner and set x-yelp-svc header for routing
    port_list = sorted(
        val["proxy_port"] for val in namespaces.values() if val.get("proxy_port")
    )
    ports = [
        k8s.V1ServicePort(
            name=f"p{port}",
            port=port,
            protocol="TCP",
            target_port=PAASTA_SVC_PORT,
            app_protocol="http",
        )
        for port in [1337, *port_list]
    ]

    service_meta = k8s.V1ObjectMeta(name=UNIFIED_K8S_SVC_NAME, annotations=ANNOTATIONS)
    service_spec = k8s.V1ServiceSpec(ports=ports)
    service_object = k8s.V1APIService(metadata=service_meta, spec=service_spec)
    yield partial(
        kube_client.core.create_namespaced_service, PAASTA_NAMESPACE, service_object
    )

    sorted_namespaces = sorted(namespaces.keys())
    x_yelp_svc_routes = [
        dict(
            match=[dict(headers={"x-yelp-svc": dict(exact=mesh_ns)})],
            delegate=dict(
                name=sanitise_kubernetes_service_name(mesh_ns),
                namespace=PAASTA_NAMESPACE,
            ),
        )
        for mesh_ns in sorted_namespaces
    ]
    port_routes = [
        dict(
            match=[dict(port=namespaces[mesh_ns]["proxy_port"])],
            delegate=dict(
                name=sanitise_kubernetes_service_name(mesh_ns),
                namespace=PAASTA_NAMESPACE,
            ),
        )
        for mesh_ns in sorted_namespaces
        if namespaces[mesh_ns].get("proxy_port")
    ]
    virtual_service = dict(
        apiVersion="networking.istio.io/v1alpha3",
        kind="VirtualService",
        metadata=dict(
            name="paasta-routing",
            namespace=PAASTA_NAMESPACE,
        ),
        spec=dict(
            hosts=["paasta-routing", "169.254.255.254"],
            http=x_yelp_svc_routes + port_routes,
        ),
    )

    yield partial(
        kube_client.custom.create_namespaced_custom_object,
        "networking.istio.io",
        "v1alpha3",
        PAASTA_NAMESPACE,
        "virtualservices",
        virtual_service,
    )


def setup_paasta_namespace_services(
    kube_client: KubeClient,
    paasta_namespaces: AbstractSet,
    existing_namespace_services: Set[str],
    existing_virtual_services: Set[str],
) -> Iterator:
    for namespace in paasta_namespaces:
        service = sanitise_kubernetes_service_name(namespace)
        if service not in existing_namespace_services:
            log.info(f"Creating k8s service {service} because it does not exist yet.")

            service_meta = k8s.V1ObjectMeta(name=service, annotations=ANNOTATIONS)
            port_spec = k8s.V1ServicePort(
                name="http", port=PAASTA_SVC_PORT, protocol="TCP", app_protocol="http"
            )
            service_spec = k8s.V1ServiceSpec(
                selector={registration_label(namespace): "true"}, ports=[port_spec]
            )
            service_object = k8s.V1APIService(metadata=service_meta, spec=service_spec)
            yield partial(
                kube_client.core.create_namespaced_service,
                PAASTA_NAMESPACE,
                service_object,
            )

        if service not in existing_virtual_services:
            log.info(
                f"Creating istio virtualservice {service} because it does not exist yet."
            )

            route = dict(
                destination=dict(host=service, port=dict(number=PAASTA_SVC_PORT))
            )
            virtual_service = dict(
                apiVersion="networking.istio.io/v1alpha3",
                kind="VirtualService",
                metadata=dict(name=service, namespace=PAASTA_NAMESPACE),
                spec=dict(http=[dict(route=[route])]),
            )
            yield partial(
                kube_client.custom.create_namespaced_custom_object,
                "networking.istio.io",
                "v1alpha3",
                PAASTA_NAMESPACE,
                "virtualservices",
                virtual_service,
            )


def cleanup_paasta_namespace_services(
    kube_client: KubeClient,
    paasta_namespaces: AbstractSet,
    existing_namespace_services: Set[str],
    existing_virtual_services: Set[str],
) -> Iterator:
    declared_services = {
        sanitise_kubernetes_service_name(ns) for ns in paasta_namespaces
    }
    for service in existing_namespace_services:
        if service == UNIFIED_K8S_SVC_NAME or service in declared_services:
            continue
        log.info(f"Garbage collecting K8s Service {service}")
        yield partial(
            kube_client.core.delete_namespaced_service, service, PAASTA_NAMESPACE
        )
    for service in existing_virtual_services:
        if service == UNIFIED_K8S_SVC_NAME or service in declared_services:
            continue
        log.info(f"Garbage collecting Istio VS {service}")
        yield partial(
            kube_client.custom.delete_namespaced_custom_object,
            "networking.istio.io",
            "v1beta1",
            PAASTA_NAMESPACE,
            "virtualservices",
            service,
        )


def process_kube_services(
    kube_client: KubeClient, soa_dir: str = DEFAULT_SOA_DIR
) -> Iterator:

    existing_namespace_services = get_existing_kubernetes_service_names(kube_client)
    existing_virtual_services = get_existing_kubernetes_virtual_services(kube_client)

    namespaces = load_smartstack_namespaces(soa_dir)

    should_setup_unified = (
        UNIFIED_K8S_SVC_NAME not in existing_namespace_services
        or UNIFIED_K8S_SVC_NAME not in existing_virtual_services
    )

    if should_setup_unified:
        log.info(f"Creating {UNIFIED_K8S_SVC_NAME} because it does not exist yet.")
        yield from setup_paasta_routing(
            kube_client=kube_client,
            namespaces=namespaces,
        )

    yield from setup_paasta_namespace_services(
        kube_client,
        namespaces.keys(),
        existing_namespace_services,
        existing_virtual_services,
    )

    yield from cleanup_paasta_namespace_services(
        kube_client,
        namespaces.keys(),
        existing_namespace_services,
        existing_virtual_services,
    )


def setup_istio_mesh(
    kube_client: KubeClient,
    rate_limit: int = 0,
    soa_dir: str = DEFAULT_SOA_DIR,
) -> bool:
    delay = 0 if rate_limit == 0 else 1.0 / float(rate_limit)
    took = delay
    success = True
    for fn in process_kube_services(kube_client=kube_client, soa_dir=soa_dir):
        time.sleep(max(0, delay - took))
        try:
            log.debug(f"Calling yielded {fn.func}({fn.args})")
            start = time.time()
            result = fn()
            took = time.time() - start
            log.debug(f"Result: {result}, took {took}s")
        except Exception:
            success = False
            log.exception(f"Failed calling {fn.func}({fn.args})")
    return success


def main() -> None:
    args = parse_args()
    logging.basicConfig(level=logging.DEBUG if args.verbose else logging.INFO)
    kube_client = KubeClient()
    ensure_namespace(kube_client, namespace=PAASTA_NAMESPACE)
    success = setup_istio_mesh(kube_client, args.rate_limit, args.soa_dir)
    sys.exit(0 if success else 1)


if __name__ == "__main__":
    main()



import contextlib
from unittest import mock
from unittest.mock import AsyncMock
from unittest.mock import MagicMock
from unittest.mock import Mock

import pytest

from paasta_tools import hacheck


@contextlib.contextmanager
def mock_ClientSession(**fake_session_kwargs):
    fake_session = MagicMock(name="session", **fake_session_kwargs)

    class FakeClientSession:
        def __init__(self, *args, **kwargs):
            ...

        async def __aenter__(*args):
            return fake_session

        async def __aexit__(*args):
            pass

    with mock.patch("aiohttp.ClientSession", new=FakeClientSession, autospec=False):
        yield


@pytest.mark.asyncio
async def test_get_spool():
    fake_response = mock.Mock(
        status=503,
        text=AsyncMock(
            return_value="Service service in down state since 1435694078.778886 "
            "until 1435694178.780000: Drained by Paasta"
        ),
    )
    fake_task = mock.Mock(host="fake_host", ports=[54321])

    with mock_ClientSession(
        get=Mock(
            return_value=MagicMock(__aenter__=AsyncMock(return_value=fake_response))
        )
    ):
        actual = await hacheck.get_spool(fake_task)

    expected = {
        "service": "service",
        "state": "down",
        "reason": "Drained by Paasta",
        "since": 1435694078.778886,
        "until": 1435694178.780000,
    }
    assert actual == expected


@pytest.mark.asyncio
async def test_get_spool_handles_no_ports():
    actual = await hacheck.get_spool(None)
    assert actual is None
